# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import copy
import pandas as pd
import numpy as np

from pandapower import create_impedance, create_switch

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)

from functools import partial
from pandapower.auxiliary import (LoadflowNotConverged, OPFNotConverged, ControllerNotConverged,
                                  NetCalculationNotConverged)
from pandapower.run import runpp
from pandapower.toolbox import get_connected_elements, replace_xward_by_ward
from pandapower.diagnostic_reports import diagnostic_report

# separator between log messages
log_message_sep = "\n --------\n"

expected_exceptions = (LoadflowNotConverged, OPFNotConverged, ControllerNotConverged, NetCalculationNotConverged)


def diagnostic(net, report_style='detailed', warnings_only=False, return_result_dict=True,
               overload_scaling_factor=0.001, min_r_ohm=0.001, min_x_ohm=0.001, max_r_ohm=100,
               max_x_ohm=100, nom_voltage_tolerance=0.3, numba_tolerance=1e-05, **kwargs):
    """
    Tool for diagnosis of pandapower networks. Identifies possible reasons for non converging loadflows.

    INPUT:
     **net** (pandapowerNet) : pandapower network

    OPTIONAL:
     - **report_style** (string, 'detailed') : style of the report, that gets ouput in the console

      'detailled': full report with high level of additional descriptions

      'compact'  : more compact report, containing essential information only

      'None'     : no report


     - **warnings_only** (boolean, False): Filters logging output for warnings

      True: logging output for errors only

      False: logging output for all checks, regardless if errors were found or not


     - **return_result_dict** (boolean, True): returns a dictionary containing all check results

      True: returns dict with all check results

      False: no result dict

     - **overload_scaling_factor** (float, 0.001): downscaling factor for loads and generation \
     for overload check

     - **lines_min_length_km** (float, 0): minimum length_km allowed for lines

     - **lines_min_z_ohm** (float, 0): minimum z_ohm allowed for lines

     - **nom_voltage_tolerance** (float, 0.3): highest allowed relative deviation between nominal \
     voltages and bus voltages

     - **kwargs** - Keyword arguments for the power flow function to use during tests. If "run" is \
     in kwargs the default call to runpp() is replaced by the function kwargs["run"]

    OUTPUT:
     - **diag_results** (dict): dict that contains the indices of all elements where errors were found

      Format: {'check_name': check_results}

    EXAMPLE:

    <<< pandapower.diagnostic(net, report_style='compact', warnings_only=True)

    """

    diag_functions = [
        (missing_bus_indices, {}),
        (disconnected_elements, {}),
        (different_voltage_levels_connected, {}),
        (implausible_impedance_values, {"min_r_ohm": min_r_ohm, "min_x_ohm": min_x_ohm, "max_r_ohm": max_r_ohm,
                                          "max_x_ohm": max_x_ohm, **kwargs}),
        (nominal_voltages_dont_match, {"nom_voltage_tolerance": nom_voltage_tolerance}),
        (invalid_values, {}),
        (overload, {"overload_scaling_factor": overload_scaling_factor, **kwargs}),
        (wrong_switch_configuration, kwargs),
        (multiple_voltage_controlling_elements_per_bus, {}),
        (no_ext_grid, {}),
        (wrong_reference_system, {}),
        (deviation_from_std_type, {}),
        (numba_comparison, {"numba_tolerance": numba_tolerance, **kwargs}),
        (parallel_switches, {}),
    ]

    diag_results = {}
    diag_errors = {}
    for diag_function, kwargs in diag_functions:
        try:
            diag_result = diag_function(net, **kwargs)
            if diag_result is not None:
                diag_results[diag_function.__name__] = diag_result
        except Exception as e:
            diag_errors[diag_function.__name__] = e

    diag_params = {
        "overload_scaling_factor": overload_scaling_factor,
        "min_r_ohm": min_r_ohm,
        "min_r_dc_ohm": min_r_ohm,
        "min_x_ohm": min_x_ohm,
        "min_r_pu": max_r_ohm,
        "min_x_pu": max_x_ohm,
        "nom_voltage_tolerance": nom_voltage_tolerance,
        "numba_tolerance": numba_tolerance
    }

    if report_style == 'detailed':
        diagnostic_report(net, diag_results, diag_errors, diag_params, compact_report=False,
                          warnings_only=warnings_only)
    elif report_style == 'compact':
        diagnostic_report(net, diag_results, diag_errors, diag_params, compact_report=True,
                          warnings_only=warnings_only)
    if return_result_dict:
        return diag_results


def check_greater_zero(element, element_index, column):
    """
     functions that check, if a certain input type restriction for attribute values of a pandapower
     elements are fulfilled. Exemplary description for all type check functions.

     INPUT:
        **element (pandas.Series)** - pandapower element instance (e.g. net.bus.loc[1])

        **element_index (int)**     - index of the element instance

        **column (string)**         - element attribute (e.g. 'vn_kv')


     OUTPUT:
        **element_index (index)**   - index of element instance, if input type restriction is not
                                      fulfilled


    """

    if check_number(element, element_index, column) is None:

        if (element[column] <= 0):
            return element_index

    else:
        return element_index


def check_greater_equal_zero(element, element_index, column):
    if check_number(element, element_index, column) is None:

        if (element[column] < 0):
            return element_index

    else:
        return element_index


def check_less_zero(element, element_index, column):
    if check_number(element, element_index, column) is None:

        if (element[column] >= 0):
            return element_index

    else:
        return element_index


def check_less_15(element, element_index, column):
    if check_number(element, element_index, column) is None:

        if element[column] >= 15:
            return element_index

    else:
        return element_index


def check_less_equal_zero(element, element_index, column):
    if check_number(element, element_index, column) is None:

        if (element[column] > 0):
            return element_index

    else:
        return element_index


def check_boolean(element, element_index, column):
    valid_values = [True, False, 0, 1, 0.0, 1.0]
    if element[column] not in valid_values:
        return element_index


def check_pos_int(element, element_index, column):
    if check_number(element, element_index, column) is None:
        if not ((element[column] % 1 == 0) and element[column] >= 0):
            return element_index

    else:
        return element_index


def check_number(element, element_index, column):
    try:
        nan_check = np.isnan(element[column])
        if nan_check or isinstance(element[column], bool):
            return element_index
    except TypeError:
        return element_index


def check_greater_zero_less_equal_one(element, element_index, column):
    if check_number(element, element_index, column) is None:

        if not (0 < element[column] <= 1):
            return element_index

    else:
        return element_index


def check_switch_type(element, element_index, column):
    valid_values = ['b', 'l', 't', 't3']
    if element[column] not in valid_values:
        return element_index


def invalid_values(net):
    """
    Applies type check functions to find violations of input type restrictions.

     INPUT:
        **net** (pandapowerNet)         - pandapower network

        **detailed_report** (boolean)   - True: detailed report of input type restriction violations
                                          False: summary only

     OUTPUT:
        **check_results** (dict)        - dict that contains all input type restriction violations
                                          grouped by element (keys)
                                          Format: {'element': [element_index, 'element_attribute',
                                                    attribute_value]}

    """

    check_results = {}

    # Contains all element attributes that are necessary to initiate a power flow calculation.
    # There's a tuple with the structure (attribute_name, input type restriction)
    # for each attribute according to pandapower data structure documantation
    # (see also type_checks function)

    important_values = {'bus': [('vn_kv', '>0'), ('in_service', 'boolean')],
                        'line': [('from_bus', 'positive_integer'),
                                 ('to_bus', 'positive_integer'),
                                 ('length_km', '>0'), ('r_ohm_per_km', '>=0'),
                                 ('x_ohm_per_km', '>=0'), ('c_nf_per_km', '>=0'),
                                 ('max_i_ka', '>0'), ('df', '0<x<=1'), ('in_service', 'boolean')],
                        'trafo': [('hv_bus', 'positive_integer'), ('lv_bus', 'positive_integer'),
                                  ('sn_mva', '>0'), ('vn_hv_kv', '>0'), ('vn_lv_kv', '>0'),
                                  ('vkr_percent', '>=0'), ('vk_percent', '>0'),
                                  ('vkr_percent', '<15'), ('vk_percent', '<15'),
                                  ('pfe_kw', '>=0'), ('i0_percent', '>=0'),
                                  ('in_service', 'boolean')],
                        'trafo3w': [('hv_bus', 'positive_integer'), ('mv_bus', 'positive_integer'),
                                    ('lv_bus', 'positive_integer'),
                                    ('sn_hv_mva', '>0'), ('sn_mv_mva', '>0'), ('sn_lv_mva', '>0'),
                                    ('vn_hv_kv', '>0'), ('vn_mv_kv', '>0'), ('vn_lv_kv', '>0'),
                                    ('vkr_hv_percent', '>=0'), ('vkr_mv_percent', '>=0'),
                                    ('vkr_lv_percent', '>=0'), ('vk_hv_percent', '>0'),
                                    ('vk_mv_percent', '>0'), ('vk_lv_percent', '>0'),
                                    ('vkr_hv_percent', '<15'), ('vkr_mv_percent', '<15'),
                                    ('vkr_lv_percent', '<15'), ('vk_hv_percent', '<15'),
                                    ('vk_mv_percent', '<15'), ('vk_lv_percent', '<15'),
                                    ('pfe_kw', '>=0'), ('i0_percent', '>=0'),
                                    ('in_service', 'boolean')],
                        'load': [('bus', 'positive_integer'), ('p_mw', 'number'),
                                 ('q_mvar', 'number'),
                                 ('scaling', '>=0'), ('in_service', 'boolean')],
                        'sgen': [('bus', 'positive_integer'), ('p_mw', 'number'),
                                 ('q_mvar', 'number'),
                                 ('scaling', '>=0'), ('in_service', 'boolean')],
                        'gen': [('bus', 'positive_integer'), ('p_mw', 'number'),
                                ('scaling', '>=0'), ('in_service', 'boolean')],
                        'ext_grid': [('bus', 'positive_integer'), ('vm_pu', '>0'),
                                     ('va_degree', 'number')],
                        'switch': [('bus', 'positive_integer'), ('element', 'positive_integer'),
                                   ('et', 'switch_type'), ('closed', 'boolean')]}

    # matches a check function to each single input type restriction
    type_checks = {'>0': check_greater_zero,
                   '>=0': check_greater_equal_zero,
                   '<0': check_less_zero,
                   '<15': check_less_15,
                   '<=0': check_less_equal_zero,
                   'boolean': check_boolean,
                   'positive_integer': check_pos_int,
                   'number': check_number,
                   '0<x<=1': check_greater_zero_less_equal_one,
                   'switch_type': check_switch_type
                   }

    for key in important_values:
        if len(net[key]) > 0:
            for value in important_values[key]:
                for i, element in net[key].iterrows():
                    check_result = type_checks[value[1]](element, i, value[0])
                    if check_result is not None:
                        if key not in check_results:
                            check_results[key] = []
                        # converts np.nan to str for easier usage of assert in pytest
                        nan_check = pd.isnull(net[key][value[0]].at[i])
                        if nan_check:
                            check_results[key].append((i, value[0],
                                                       str(net[key][value[0]].at[i]), value[1]))
                        else:
                            check_results[key].append((i, value[0],
                                                       net[key][value[0]].at[i], value[1]))
    if check_results:
        return check_results


def no_ext_grid(net):
    """
    Checks, if at least one external grid exists.

     INPUT:
        **net** (pandapowerNet)         - pandapower network

    """

    if net.ext_grid.in_service.sum() + (net.gen.slack & net.gen.in_service).sum() == 0:
        return True


def multiple_voltage_controlling_elements_per_bus(net):
    """
    Checks, if there are buses with more than one generator and/or more than one external grid.

     INPUT:
        **net** (pandapowerNet)         - pandapower network

        **detailed_report** (boolean)   - True: detailed report of errors found
                                      l    False: summary only

     OUTPUT:
        **check_results** (dict)        - dict that contains all buses with multiple generator and
                                          all buses with multiple external grids
                                          Format: {'mult_ext_grids': [buses]
                                                   'buses_with_mult_gens', [buses]}

    """
    check_results = {}
    buses_with_mult_ext_grids = list(net.ext_grid.groupby("bus").count().query("vm_pu > 1").index)
    if buses_with_mult_ext_grids:
        check_results['buses_with_mult_ext_grids'] = buses_with_mult_ext_grids
    buses_with_gens_and_ext_grids = set(net.ext_grid.bus).intersection(set(net.gen.bus))
    if buses_with_gens_and_ext_grids:
        check_results['buses_with_gens_and_ext_grids'] = list(buses_with_gens_and_ext_grids)

    if check_results:
        return check_results


def overload(net, overload_scaling_factor, **kwargs):
    """
    Checks, if a loadflow calculation converges. If not, checks, if an overload is the reason for
    that by scaling down the loads, gens and sgens to 0.1%.

     INPUT:
        **net** (pandapowerNet)         - pandapower network

        **kwargs** - Keyword arguments for power flow function. If "run" is in kwargs the default call to runpp()
        is replaced by the function kwargs["run"]


     OUTPUT:
        **check_results** (dict)        - dict with the results of the overload check
                                          Format: {'load_overload': True/False
                                                   'generation_overload', True/False}

    """
    # get function to run power flow
    run = partial(kwargs.pop("run", runpp), **kwargs)
    check_result = {}
    load_scaling = copy.deepcopy(net.load.scaling)
    gen_scaling = copy.deepcopy(net.gen.scaling)
    sgen_scaling = copy.deepcopy(net.sgen.scaling)

    try:
        run(net)
    except expected_exceptions:
        check_result['load'] = False
        check_result['generation'] = False
        try:
            net.load.scaling = overload_scaling_factor
            run(net)
            check_result['load'] = True
        except expected_exceptions:
            net.load.scaling = load_scaling
            try:
                net.gen.scaling = overload_scaling_factor
                net.sgen.scaling = overload_scaling_factor
                run(net)
                check_result['generation'] = True
            except expected_exceptions:
                net.sgen.scaling = sgen_scaling
                net.gen.scaling = gen_scaling
                try:
                    net.load.scaling = overload_scaling_factor
                    net.gen.scaling = overload_scaling_factor
                    net.sgen.scaling = overload_scaling_factor
                    run(net)
                    check_result['generation'] = True
                    check_result['load'] = True
                except expected_exceptions:
                    logger.debug("Overload check did not help")
        net.sgen.scaling = sgen_scaling
        net.gen.scaling = gen_scaling
        net.load.scaling = load_scaling
    except Exception as e:
        logger.error(f"Overload check failed: {str(e)}")
    if check_result:
        return check_result


def wrong_switch_configuration(net, **kwargs):
    """
    Checks, if a loadflow calculation converges. If not, checks, if the switch configuration is
    the reason for that by closing all switches

     INPUT:
        **net** (pandapowerNet)         - pandapower network

        **kwargs** - Keyword arguments for power flow function. If "run" is in kwargs the default call to runpp()
        is replaced by the function kwargs["run"]

     OUTPUT:
        **check_result** (boolean)

    """
    run = partial(kwargs.pop("run", runpp), **kwargs)
    switch_configuration = copy.deepcopy(net.switch.closed)
    try:
        run(net)
    except expected_exceptions:
        try:
            net.switch.closed = True
            run(net)
            net.switch.closed = switch_configuration
            return True
        except expected_exceptions:
            net.switch.closed = switch_configuration
            return False
    except Exception as e:
        logger.error(f"Switch check failed: {str(e)}")


def missing_bus_indices(net):
    """
        Checks for missing bus indices.

         INPUT:
            **net** (PandapowerNet)    - pandapower network


         OUTPUT:
            **check_results** (list)   - List of tuples each containing missing bus indices.
                                         Format:
                                         [(element_index, bus_name (e.g. "from_bus"),  bus_index]

    """
    check_results = {}
    bus_indices = set(net.bus.index)
    element_bus_names = {"ext_grid": ["bus"], "load": ["bus"], "gen": ["bus"], "sgen": ["bus"],
                         "trafo": ["lv_bus", "hv_bus"], "trafo3w": ["lv_bus", "mv_bus", "hv_bus"],
                         "switch": ["bus", "element"], "line": ["from_bus", "to_bus"]}
    for element in element_bus_names.keys():
        element_check = []
        for i, row in net[element].iterrows():
            for bus_name in element_bus_names[element]:
                if row[bus_name] not in bus_indices:
                    if not ((element == "switch") and (bus_name == "element") and (row.et in ['l', 't', 't3'])):
                        element_check.append((i, bus_name, row[bus_name]))
        if element_check:
            check_results[element] = element_check

    if check_results:
        return check_results


def different_voltage_levels_connected(net):
    """
    Checks, if there are lines or switches that connect different voltage levels.

     INPUT:
        **net** (pandapowerNet)         - pandapower network

     OUTPUT:
        **check_results** (dict)        - dict that contains all lines and switches that connect
                                          different voltage levels.
                                          Format: {'lines': lines, 'switches': switches}

    """
    check_results = {}
    inconsistent_lines = []
    for i, line in net.line.iterrows():
        buses = net.bus.loc[[line.from_bus, line.to_bus]]
        if buses.vn_kv.iloc[0] != buses.vn_kv.iloc[1]:
            inconsistent_lines.append(i)

    inconsistent_switches = []
    for i, switch in net.switch[net.switch.et == "b"].iterrows():
        buses = net.bus.loc[[switch.bus, switch.element]]
        if buses.vn_kv.iloc[0] != buses.vn_kv.iloc[1]:
            inconsistent_switches.append(i)

    if inconsistent_lines:
        check_results['lines'] = inconsistent_lines
    if inconsistent_switches:
        check_results['switches'] = inconsistent_switches
    if check_results:
        return check_results


def implausible_impedance_values(net, min_r_ohm, min_x_ohm, max_r_ohm, max_x_ohm, **kwargs):
    """
    Checks, if there are lines, xwards or impedances with an impedance value close to zero.

     INPUT:
        **net** (pandapowerNet)         - pandapower network

        **kwargs** - Keyword arguments for power flow function. If "run" is in kwargs the default call to runpp()
        is replaced by the function kwargs["run"]

     OUTPUT:
        **implausible_lines** (list)    - list that contains the indices of all lines with an
                                          impedance value of zero.


    """
    # get function to run power flow
    run = partial(kwargs.pop("run", runpp), **kwargs)
    check_results = []
    implausible_elements = {}

    line = net.line.loc[((net.line.r_ohm_per_km * net.line.length_km >= max_r_ohm) |
                         (net.line.r_ohm_per_km * net.line.length_km <= min_r_ohm) |
                         (net.line.x_ohm_per_km * net.line.length_km >= max_x_ohm) |
                         (net.line.x_ohm_per_km * net.line.length_km <= min_x_ohm)) &
                        net.line.in_service].index

    xward = net.xward.loc[((net.xward.r_ohm.abs() >= max_r_ohm) |
                           (net.xward.r_ohm.abs() <= min_r_ohm) |
                           (net.xward.x_ohm.abs() >= max_x_ohm) |
                           (net.xward.x_ohm.abs() <= min_x_ohm)) &
                          net.xward.in_service].index

    zb_f_ohm = np.square(net.bus.loc[net.impedance.from_bus.values, "vn_kv"].values) / net.impedance.sn_mva
    zb_t_ohm = np.square(net.bus.loc[net.impedance.to_bus.values, "vn_kv"].values) / net.impedance.sn_mva
    impedance = net.impedance.loc[((np.abs(net.impedance.rft_pu) >= max_r_ohm / zb_f_ohm) |
                                   (np.abs(net.impedance.rft_pu) <= min_r_ohm / zb_f_ohm) |
                                   (np.abs(net.impedance.xft_pu) >= max_x_ohm / zb_f_ohm) |
                                   (np.abs(net.impedance.xft_pu) <= min_x_ohm / zb_f_ohm) |
                                   (np.abs(net.impedance.rtf_pu) >= max_r_ohm / zb_t_ohm) |
                                   (np.abs(net.impedance.rtf_pu) <= min_r_ohm / zb_t_ohm) |
                                   (np.abs(net.impedance.xtf_pu) >= max_x_ohm / zb_t_ohm) |
                                   (np.abs(net.impedance.xtf_pu) <= min_x_ohm / zb_t_ohm)) &
                                  net.impedance.in_service].index

    trafo = net.trafo.loc[
        (((net.trafo.vk_percent / 100 * np.square(net.trafo.vn_hv_kv) / net.trafo.sn_mva >= max_x_ohm) |
          (net.trafo.vk_percent / 100 * np.square(net.trafo.vn_lv_kv) / net.trafo.sn_mva <= min_x_ohm)) &
         net.trafo.in_service)].index

    trafo3w = net.trafo3w.loc[
        (((net.trafo3w.vk_hv_percent / 100 * np.square(net.trafo3w.vn_hv_kv) / net.trafo3w.sn_hv_mva >= max_x_ohm) |
          (net.trafo3w.vk_hv_percent / 100 * np.square(net.trafo3w.vn_mv_kv) / net.trafo3w.sn_hv_mva <= min_x_ohm) |
          (net.trafo3w.vk_mv_percent / 100 * np.square(net.trafo3w.vn_mv_kv) / net.trafo3w.sn_mv_mva >= max_x_ohm) |
          (net.trafo3w.vk_mv_percent / 100 * np.square(net.trafo3w.vn_lv_kv) / net.trafo3w.sn_mv_mva <= min_x_ohm) |
          (net.trafo3w.vk_lv_percent / 100 * np.square(net.trafo3w.vn_hv_kv) / net.trafo3w.sn_lv_mva >= max_x_ohm) |
          (net.trafo3w.vk_lv_percent / 100 * np.square(net.trafo3w.vn_lv_kv) / net.trafo3w.sn_lv_mva <= min_x_ohm)) &
         net.trafo3w.in_service)].index
    vsc = net.vsc.loc[((net.vsc.r_ohm <= min_r_ohm)
                       | (net.vsc.x_ohm <= min_x_ohm)
                       | (net.vsc.r_dc_ohm <= min_r_ohm)) & net.vsc.in_service].index

    line_dc = net.line_dc.loc[((net.line_dc.r_ohm_per_km * net.line_dc.length_km) <= min_r_ohm) &
                              net.line_dc.in_service].index
    if len(line) > 0:
        implausible_elements['line'] = list(line)
    if len(xward) > 0:
        implausible_elements['xward'] = list(xward)
    if len(impedance) > 0:
        implausible_elements['impedance'] = list(impedance)
    if len(trafo) > 0:
        implausible_elements['trafo'] = list(trafo)
    if len(trafo3w) > 0:
        implausible_elements['trafo3w'] = list(trafo3w)
    if len(vsc) > 0:
        implausible_elements['vsc'] = list(vsc)
    if len(line_dc) > 0:
        implausible_elements['line_dc'] = list(line_dc)
    check_results.append(implausible_elements)
    # checks if loadflow converges when implausible lines or impedances are replaced by switches
    if ("line" in implausible_elements) or ("impedance" in implausible_elements) or ("xward" in implausible_elements):
        switch_copy = copy.deepcopy(net.switch)
        line_copy = copy.deepcopy(net.line)
        impedance_copy = copy.deepcopy(net.impedance)
        vsc_copy = copy.deepcopy(net.vsc)
        line_dc_copy = copy.deepcopy(net.line_dc)
        ward_copy = copy.deepcopy(net.ward)
        xward_copy = copy.deepcopy(net.xward)
        trafo_copy = copy.deepcopy(net.trafo)
        trafo3w_copy = copy.deepcopy(net.trafo3w)
        try:
            run(net)
        except (*expected_exceptions, FloatingPointError):
            try:
                for key in implausible_elements:
                    implausible_idx = implausible_elements[key]
                    if key == 'vsc':
                        net.vsc.x_ohm = np.fmax(net.vsc.x_ohm, 0.5)
                        net.vsc.r_dc_ohm = np.fmax(net.vsc.r_dc_ohm, 0.5)
                    if key == 'line_dc':
                        net.line_dc.length_km = np.fmax(net.line_dc.length_km, 0.5)
                        net.line_dc.r_dc_ohm = np.fmax(net.line_dc.r_dc_ohm, 0.5)
                    net[key].loc[implausible_idx, "in_service"] = False
                    if key == 'xward':
                        replace_xward_by_ward(net, implausible_idx)
                    elif key == 'trafo':
                        for idx in implausible_idx:
                            create_impedance(net, net.trafo.at[idx, "hv_bus"], net.trafo.at[idx, "lv_bus"],
                                                0, 0.01, 100)
                    elif key == 'trafo3w':
                        for idx in implausible_idx:
                            create_impedance(net, net.trafo3w.at[idx, "hv_bus"], net.trafo3w.at[idx, "mv_bus"],
                                                0, 0.01, 100)
                            create_impedance(net, net.trafo3w.at[idx, "mv_bus"], net.trafo3w.at[idx, "lv_bus"],
                                                0, 0.01, 100)
                            create_impedance(net, net.trafo3w.at[idx, "hv_bus"], net.trafo3w.at[idx, "lv_bus"],
                                                0, 0.01, 100)
                    else:
                        for idx in implausible_idx:
                            create_switch(net, net[key].from_bus.at[idx], net[key].to_bus.at[idx], et="b")
                run(net)
                switch_replacement = True
            except expected_exceptions:
                switch_replacement = False
            check_results.append({"loadflow_converges_with_switch_replacement": switch_replacement})
        except Exception as e:
            logger.error(f"Impedance values check failed: {str(e)}")
        net.switch = switch_copy
        net.line = line_copy
        net.impedance = impedance_copy
        net.vsc = vsc_copy
        net.line_dc = line_dc_copy
        net.ward = ward_copy
        net.xward = xward_copy
        net.trafo = trafo_copy
        net.trafo3w = trafo3w_copy
    if implausible_elements:
        return check_results


def nominal_voltages_dont_match(net, nom_voltage_tolerance):
    """
    Checks, if there are components whose nominal voltages differ from the nominal voltages of the
    buses they're connected to. At the moment, only trafos and trafo3w are checked.
    Also checks for trafos with swapped hv and lv connectors.

     INPUT:
        **net** (pandapowerNet)         - pandapower network

     OUTPUT:
        **check_results** (dict)        - dict that contains all components whose nominal voltages
                                          differ from the nominal voltages of the buses they're
                                          connected to.

                                          Format:

                                          {trafo': {'hv_bus' : trafos_indices,
                                                    'lv_bus' : trafo_indices,
                                                    'hv_lv_swapped' : trafo_indices},
                                           trafo3w': {'hv_bus' : trafos3w_indices,
                                                      'mv_bus' : trafos3w_indices
                                                      'lv_bus' : trafo3w_indices,
                                                      'connectors_swapped_3w' : trafo3w_indices}}

    """
    results = {}
    trafo_results = {}
    trafo3w_results = {}

    hv_bus = []
    lv_bus = []
    hv_lv_swapped = []

    hv_bus_3w = []
    mv_bus_3w = []
    lv_bus_3w = []
    connectors_swapped_3w = []

    for i, trafo in net.trafo.iterrows():
        hv_bus_violation = False
        lv_bus_violation = False
        connectors_swapped = False
        hv_bus_vn_kv = net.bus.vn_kv.at[trafo.hv_bus]
        lv_bus_vn_kv = net.bus.vn_kv.at[trafo.lv_bus]

        if abs(1 - (trafo.vn_hv_kv / hv_bus_vn_kv)) > nom_voltage_tolerance:
            hv_bus_violation = True
        if abs(1 - (trafo.vn_lv_kv / lv_bus_vn_kv)) > nom_voltage_tolerance:
            lv_bus_violation = True
        if hv_bus_violation and lv_bus_violation:
            trafo_voltages = np.array(([trafo.vn_hv_kv, trafo.vn_lv_kv]))
            bus_voltages = np.array([hv_bus_vn_kv, lv_bus_vn_kv])
            trafo_voltages.sort()
            bus_voltages.sort()
            if all((abs(trafo_voltages - bus_voltages) / bus_voltages) < (nom_voltage_tolerance)):
                connectors_swapped = True

        if connectors_swapped:
            hv_lv_swapped.append(i)
        else:
            if hv_bus_violation:
                hv_bus.append(i)
            if lv_bus_violation:
                lv_bus.append(i)

    if hv_bus:
        trafo_results['hv_bus'] = hv_bus
    if lv_bus:
        trafo_results['lv_bus'] = lv_bus
    if hv_lv_swapped:
        trafo_results['hv_lv_swapped'] = hv_lv_swapped
    if trafo_results:
        results['trafo'] = trafo_results

    for i, trafo3w in net.trafo3w.iterrows():
        hv_bus_violation = False
        mv_bus_violation = False
        lv_bus_violation = False
        connectors_swapped = False
        hv_bus_vn_kv = net.bus.vn_kv.at[trafo3w.hv_bus]
        mv_bus_vn_kv = net.bus.vn_kv.at[trafo3w.mv_bus]
        lv_bus_vn_kv = net.bus.vn_kv.at[trafo3w.lv_bus]

        if abs(1 - (trafo3w.vn_hv_kv / hv_bus_vn_kv)) > nom_voltage_tolerance:
            hv_bus_violation = True
        if abs(1 - (trafo3w.vn_mv_kv / mv_bus_vn_kv)) > nom_voltage_tolerance:
            mv_bus_violation = True
        if abs(1 - (trafo3w.vn_lv_kv / lv_bus_vn_kv)) > nom_voltage_tolerance:
            lv_bus_violation = True
        if hv_bus_violation and mv_bus_violation and lv_bus_violation:
            trafo_voltages = np.array(([trafo3w.vn_hv_kv, trafo3w.vn_mv_kv, trafo3w.vn_lv_kv]))
            bus_voltages = np.array([hv_bus_vn_kv, mv_bus_vn_kv, lv_bus_vn_kv])
            trafo_voltages.sort()
            bus_voltages.sort()
            if all((abs(trafo_voltages - bus_voltages) / bus_voltages) < (nom_voltage_tolerance)):
                connectors_swapped = True

        if connectors_swapped:
            connectors_swapped_3w.append(i)
        else:
            if hv_bus_violation:
                hv_bus_3w.append(i)
            if mv_bus_violation:
                mv_bus_3w.append(i)
            if lv_bus_violation:
                lv_bus_3w.append(i)

    if hv_bus_3w:
        trafo3w_results['hv_bus'] = hv_bus_3w
    if mv_bus_3w:
        trafo3w_results['mv_bus'] = mv_bus_3w
    if lv_bus_3w:
        trafo3w_results['lv_bus'] = lv_bus_3w
    if connectors_swapped_3w:
        trafo3w_results['connectors_swapped_3w'] = connectors_swapped_3w
    if trafo3w_results:
        results['trafo3w'] = trafo3w_results

    if len(results) > 0:
        return results


def disconnected_elements(net):
    """
    Checks, if there are network sections without a connection to an ext_grid. Returns all network
    elements in these sections, that are in service. Elements belonging to the same disconnected
    networks section are grouped in lists (e.g. disconnected lines: [[1, 2, 3], [4, 5]]
    means, that lines 1, 2 and 3 are in one disconncted section but are connected to each other.
    The same stands for lines 4, 5.)

     INPUT:
        **net** (pandapowerNet)         - pandapower network

     OUTPUT:
        **disc_elements** (dict)        - list that contains all network elements, without a
                                          connection to an ext_grid.

                                          format: {'disconnected buses'   : bus_indices,
                                                   'disconnected switches' : switch_indices,
                                                   'disconnected lines'    : line_indices,
                                                   'disconnected trafos'   : trafo_indices
                                                   'disconnected loads'    : load_indices,
                                                   'disconnected gens'     : gen_indices,
                                                   'disconnected sgens'    : sgen_indices}

    """
    from pandapower.topology import create_nxgraph, connected_components
    mg = create_nxgraph(net)
    sections = connected_components(mg)
    disc_elements = []

    for section in sections:
        section_dict = {}

        if not section & set(net.ext_grid.bus[net.ext_grid.in_service]).union(
                net.gen.bus[net.gen.slack & net.gen.in_service]) and any(
                net.bus.in_service.loc[list(section)]):
            section_buses = list(net.bus[net.bus.index.isin(section) & net.bus.in_service].index)
            section_switches = list(net.switch[net.switch.bus.isin(section_buses)].index)
            section_lines = list(get_connected_elements(net, 'line', section_buses,
                                                        respect_switches=True,
                                                        respect_in_service=True))
            section_trafos = list(get_connected_elements(net, 'trafo', section_buses,
                                                         respect_switches=True,
                                                         respect_in_service=True))

            section_trafos3w = list(get_connected_elements(net, 'trafo3w', section_buses,
                                                           respect_switches=True,
                                                           respect_in_service=True))
            section_gens = list(net.gen[net.gen.bus.isin(section) & net.gen.in_service].index)
            section_sgens = list(net.sgen[net.sgen.bus.isin(section) & net.sgen.in_service].index)
            section_loads = list(net.load[net.load.bus.isin(section) & net.load.in_service].index)

            if section_buses:
                section_dict['buses'] = section_buses
            if section_switches:
                section_dict['switches'] = section_switches
            if section_lines:
                section_dict['lines'] = section_lines
            if section_trafos:
                section_dict['trafos'] = section_trafos
            if section_trafos3w:
                section_dict['trafos3w'] = section_trafos3w
            if section_loads:
                section_dict['loads'] = section_loads
            if section_gens:
                section_dict['gens'] = section_gens
            if section_sgens:
                section_dict['sgens'] = section_sgens

            if any(section_dict.values()):
                disc_elements.append(section_dict)

    open_trafo_switches = net.switch[(net.switch.et == 't') & (net.switch.closed == 0)]
    isolated_trafos = set(
        (open_trafo_switches.groupby("element").count().query("bus > 1").index))
    isolated_trafos_is = isolated_trafos.intersection((set(net.trafo[net.trafo.in_service].index)))
    if isolated_trafos_is:
        disc_elements.append({'isolated_trafos': list(isolated_trafos_is)})

    isolated_trafos3w = set(open_trafo_switches.groupby("element").count().query("bus > 2").index)
    isolated_trafos3w_is = isolated_trafos3w.intersection(set(net.trafo[net.trafo.in_service].index))
    if isolated_trafos3w_is:
        disc_elements.append({'isolated_trafos3w': list(isolated_trafos3w_is)})

    if disc_elements:
        return disc_elements


def wrong_reference_system(net):
    """
    Checks usage of wrong reference system for loads, sgens and gens.

     INPUT:
        **net** (pandapowerNet)    - pandapower network

     OUTPUT:
        **check_results** (dict)        - dict that contains the indices of all components where the
                                          usage of the wrong reference system was found.

                                          Format: {'element_type': element_indices}

    """
    check_results = {}
    neg_loads = list(net.load[net.load.p_mw < 0].index)
    neg_gens = list(net.gen[net.gen.p_mw < 0].index)
    neg_sgens = list(net.sgen[net.sgen.p_mw < 0].index)

    if neg_loads:
        check_results['loads'] = neg_loads
    if neg_gens:
        check_results['gens'] = neg_gens
    if neg_sgens:
        check_results['sgens'] = neg_sgens

    if check_results:
        return check_results


def numba_comparison(net, numba_tolerance, **kwargs):
    """
        Compares the results of loadflows with numba=True vs. numba=False.

         INPUT:
            **net** (pandapowerNet)    - pandapower network
            **numba_tolerance** (float) - Maximum absolute deviation allowed between
                                          numba=True/False results.
         OPTIONAL:
            **kwargs** - Keyword arguments for power flow function. If "run" is in kwargs the default call to runpp()
                         is replaced by the function kwargs["run"]
         OUTPUT:
            **check_result** (dict)    - Absolute deviations between numba=True/False results.
    """
    run = partial(kwargs.pop("run", runpp), **kwargs)
    check_results = {}
    run(net, numba=True)
    result_numba_true = copy.deepcopy(net)
    run(net, numba=False)
    result_numba_false = copy.deepcopy(net)
    res_keys = [key for key in result_numba_true.keys() if
                (key in ['res_bus', 'res_ext_grid',
                         'res_gen', 'res_impedance',
                         'res_line', 'res_load',
                         'res_sgen', 'res_shunt',
                         'res_trafo', 'res_trafo3w',
                         'res_ward', 'res_xward'])]
    for key in res_keys:
        diffs = abs(result_numba_true[key] - result_numba_false[key]) > numba_tolerance
        if any(diffs.any()):
            if (key not in check_results.keys()):
                check_results[key] = {}
            for col in diffs.columns:
                if (col not in check_results[key].keys()) and (diffs.any()[col]):
                    check_results[key][col] = {}
                    numba_true = result_numba_true[key][col][diffs[col]]
                    numba_false = result_numba_false[key][col][diffs[col]]
                    check_results[key][col] = abs(numba_true - numba_false)

    if check_results:
        return check_results


def deviation_from_std_type(net):
    """
        Checks, if element parameters match the values in the standard type library.

         INPUT:
            **net** (pandapowerNet)    - pandapower network


         OUTPUT:
            **check_results** (dict)   - All elements, that don't match the values in the
                                         standard type library

                                         Format: (element_type, element_index, parameter)


    """
    check_results = {}
    for key in net.std_types.keys():
        if key in net:
            for i, element in net[key].iterrows():
                std_type = element.std_type
                if std_type in net.std_types[key].keys():
                    std_type_values = net.std_types[key][std_type]
                    for param in std_type_values.keys():
                        if param == "tap_pos":
                            continue
                        if param in net[key].columns:
                            try:
                                isclose = np.isclose(element[param], std_type_values[param],
                                                     equal_nan=True)
                            except TypeError:
                                isclose = element[param] == std_type_values[param]
                            if not isclose:
                                if key not in check_results.keys():
                                    check_results[key] = {}
                                check_results[key][i] = {'param': param, 'e_value': element[param],
                                                         'std_type_value': std_type_values[param],
                                                         'std_type_in_lib': True}
                elif std_type is not None:
                    if key not in check_results.keys():
                        check_results[key] = {}
                    check_results[key][i] = {'std_type_in_lib': False}

    if check_results:
        return check_results


def parallel_switches(net):
    """
    Checks for parallel switches.

     INPUT:
        **net** (PandapowerNet)    - pandapower network

     OUTPUT:
        **parallel_switches** (list)   - List of tuples each containing parallel switches.
    """
    parallel_switches = []
    compare_parameters = ['bus', 'element', 'et']
    parallels_bus_and_element = list(
        net.switch.groupby(compare_parameters).count().query('closed > 1').index)
    for bus, element, et in parallels_bus_and_element:
        parallel_switches.append(list(net.switch.query('bus==@bus & element==@element & et==@et').index))
    if parallel_switches:
        return parallel_switches



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

@author: mrichter
"""
from __future__ import annotations
import datetime
import enum
import json
from typing import List, Dict
from . import interfaces


class ReportCode(enum.Enum):
    INFO = 20
    INFO_PARSING = 21
    INFO_CONVERTING = 22
    INFO_REPAIRING = 23
    WARNING = 30
    WARNING_PARSING = 31
    WARNING_CONVERTING = 32
    WARNING_REPAIRING = 33
    ERROR = 40
    ERROR_PARSING = 41
    ERROR_CONVERTING = 42
    ERROR_REPAIRING = 43
    EXCEPTION = 50
    EXCEPTION_PARSING = 51
    EXCEPTION_CONVERTING = 52
    EXCEPTION_REPAIRING = 53

    @classmethod
    def from_value(cls, value):
        for member in cls:
            if member.value == value:
                return member
        raise ValueError(f"{value} is not a valid value for ReportCode")


class LogLevel(enum.Enum):
    INFO = 'INFO'
    WARNING = 'WARNING'
    ERROR = 'ERROR'
    EXCEPTION = 'EXCEPTION'

    @classmethod
    def from_value(cls, value):
        for member in cls:
            if member.value == value:
                return member
        raise ValueError(f"{value} is not a valid value for LogLevel")


class Report:
    def __init__(self, message: str, code: ReportCode, level: LogLevel, timestamp: datetime.datetime = None):
        self.timestamp = datetime.datetime.now()
        if timestamp is not None:
            self.timestamp = timestamp
        self.message = message
        self.code = code
        self.level = level

    def to_dict(self) -> Dict:
        return {
            "timestamp": self.timestamp.isoformat(),
            "message": self.message,
            "code": self.code.value,
            "level": self.level.value
        }


class ReportContainer(interfaces.ReportContainer):
    def __init__(self, logs: List[Report] = None):
        if logs is not None and isinstance(logs, List):
            self.logs = logs
        else:
            self.logs = []
    
    def add_log(self, log: Report):
        self.logs.append(log)

    def get_logs(self) -> List[Report]:
        return self.logs

    def extend_log(self, report_container: ReportContainer):
        self.logs.extend(report_container.get_logs())

    def serialize(self, path_to_disk: str):
        with open(path_to_disk, 'w', encoding='UTF-8') as file:
            json.dump(obj=[log.to_dict() for log in self.logs], fp=file, indent=2)



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np
from scipy.sparse import csr_matrix, vstack, hstack
from scipy.sparse.linalg import spsolve, norm, inv

from pandapower.estimation.algorithm.estimator import BaseEstimatorIRWLS, get_estimator
from pandapower.estimation.algorithm.matrix_base import BaseAlgebra, \
    BaseAlgebraZeroInjConstraints
from pandapower.estimation.idx_bus import ZERO_INJ_FLAG, P, P_STD, Q, Q_STD
from pandapower.estimation.ppc_conversion import ExtendedPPCI
from pandapower.pypower.idx_bus import bus_cols

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging
std_logger = logging.getLogger(__name__)

__all__ = ["WLSAlgorithm", "WLSZeroInjectionConstraintsAlgorithm", "IRWLSAlgorithm"]


class BaseAlgorithm:
    def __init__(self, tolerance, maximum_iterations, logger=std_logger):
        self.tolerance = tolerance
        self.max_iterations = maximum_iterations
        self.logger = logger
        self.successful = False
        self.iterations = None

        # Parameters for estimate
        self.eppci = None
        self.pp_meas_indices = None

    def check_observability(self, eppci: ExtendedPPCI, z):
        # Check if observability criterion is fulfilled and the state estimation is possible
        num_slacks = sum(~eppci.non_slack_bus_mask)
        if len(z) < 2 * eppci["bus"].shape[0] - num_slacks:
            self.logger.error("System is not observable (cancelling)")
            self.logger.error("Measurements available: %d. Measurements required: %d" %
                              (len(z), 2 * eppci["bus"].shape[0] - 1))
            raise UserWarning("Measurements available: %d. Measurements required: %d" %
                              (len(z), 2 * eppci["bus"].shape[0] - 1))

    def check_result(self, current_error, cur_it):
        # print output for results
        if current_error <= self.tolerance:
            self.successful = True
            self.logger.debug("State Estimation successful ({:d} iterations)".format(cur_it))
        else:
            self.successful = False
            self.logger.debug("State Estimation not successful ({:d}/{:d} iterations)".format(cur_it,
                                                                                              self.max_iterations))

    def initialize(self, eppci: ExtendedPPCI):
        # Check observability
        self.eppci = eppci
        self.pp_meas_indices = eppci.pp_meas_indices
        self.check_observability(eppci, eppci.z)

    def estimate(self, eppci: ExtendedPPCI, **kwargs):
        # Must be implemented individually!!
        pass


class WLSAlgorithm(BaseAlgorithm):
    def __init__(self, tolerance, maximum_iterations, logger=std_logger):
        super(WLSAlgorithm, self).__init__(tolerance, maximum_iterations, logger)

        # Parameters for Bad data detection
        self.R_inv = None
        self.Gm = None
        self.r = None
        self.H = None
        self.hx = None
        self.iterations = None
        self.obj_func = None
        logging.basicConfig(level=logging.DEBUG)

    def estimate(self, eppci: ExtendedPPCI, **kwargs):
        self.initialize(eppci)
        # matrix calculation object
        sem = BaseAlgebra(eppci)

        current_error, cur_it = 100., 0
        # invert covariance matrix
        eppci.r_cov[eppci.r_cov<(10**(-5))] = 10**(-5)
        r_inv = csr_matrix(np.diagflat(1 / eppci.r_cov ** 2))
        E = eppci.E
        while current_error > self.tolerance and cur_it < self.max_iterations:
            # self.logger.debug("Starting iteration {:d}".format(1 + cur_it))
            try:
                # residual r
                r = csr_matrix(sem.create_rx(E)).T

                # jacobian matrix H
                H = csr_matrix(sem.create_hx_jacobian(E))

                # remove current magnitude measurements at the first iteration 
                # because with flat start they have null derivative
                if cur_it == 0 and eppci.any_i_meas:
                    idx = eppci.idx_non_imeas
                    r_inv = r_inv[idx,:][:,idx]
                    r = r[idx,:]
                    H = H[idx,:]

                # gain matrix G_m
                # G_m = H^t * R^-1 * H
                G_m = H.T * (r_inv * H)
                norm_G = norm(G_m, np.inf)
                norm_invG = norm(inv(G_m), np.inf)
                cond = norm_G*norm_invG
                if cond > 10**18:
                    self.logger.warning("WARNING: Gain matrix is ill-conditioned: {:.2E}".format(cond))

                # state vector difference d_E
                # d_E = G_m^-1 * (H' * R^-1 * r)
                d_E = spsolve(G_m, H.T * (r_inv * r))

                # Scaling of Delta_X to avoid divergence due o ill-conditioning and 
                # operating conditions far from starting state variables
                current_error = np.max(np.abs(d_E))
                if current_error > 0.35:
                    d_E = d_E*0.35/current_error

                # Update E with d_E
                E += d_E.ravel()
                eppci.update_E(E)

                # log data 
                # obj_func = (r.T*r_inv*r)[0,0]
                # self.logger.debug("Current delta_x: {:.7f}".format(current_error))
                # self.logger.debug("Current objective function value: {:.1f}".format(obj_func))

                # Restore full weighting matrix with current measurements
                if cur_it == 0 and eppci.any_i_meas:
                    r_inv = csr_matrix(np.diagflat(1 / eppci.r_cov ** 2))

                # prepare next iteration
                cur_it += 1

            except np.linalg.linalg.LinAlgError:
                self.logger.error("A problem appeared while using the linear algebra methods."
                                  "Check and change the measurement set.")
                return False

        # check if the estimation is successfull
        self.check_result(current_error, cur_it)
        self.iterations = cur_it
        # self.obj_func = obj_func
        if self.successful:
            # store variables required for chi^2 and r_N_max test:
            self.R_inv = r_inv.toarray()
            self.Gm = G_m.toarray()
            self.r = r.toarray()
            self.H = H.toarray()
            # create h(x) for the current iteration
            self.hx = sem.create_hx(eppci.E)
        return eppci


class WLSZeroInjectionConstraintsAlgorithm(BaseAlgorithm):
    def estimate(self, eppci: ExtendedPPCI, **kwargs):
        # state vector built from delta, |V| and zero injections
        # Find pq bus with zero p,q and shunt admittance
        if not np.any(eppci["bus"][:, bus_cols + ZERO_INJ_FLAG]):
            raise UserWarning("Network has no bus with zero injections! Please use WLS instead!")
        zero_injection_bus = np.argwhere(eppci["bus"][:, bus_cols + ZERO_INJ_FLAG]).ravel()
        eppci["bus"][np.ix_(zero_injection_bus, [bus_cols + P, bus_cols + P_STD, bus_cols + Q, bus_cols + Q_STD])] = np.nan
        # Withn pq buses with zero injection identify those who have also no p or q measurement
        p_zero_injections = zero_injection_bus
        q_zero_injections = zero_injection_bus
        new_states = np.zeros(len(p_zero_injections) + len(q_zero_injections))

        num_bus = eppci["bus"].shape[0]

        # matrix calculation object
        sem = BaseAlgebraZeroInjConstraints(eppci)

        current_error, cur_it = 100., 0
        r_inv = csr_matrix((np.diagflat(1 / eppci.r_cov) ** 2))
        E = eppci.E
        # update the E matrix
        E_ext = np.r_[eppci.E, new_states]

        while current_error > self.tolerance and cur_it < self.max_iterations:
            self.logger.debug("Starting iteration {:d}".format(1 + cur_it))
            try:
                c_x = sem.create_cx(E, p_zero_injections, q_zero_injections)

                # residual r
                r = csr_matrix(sem.create_rx(E)).T
                c_rxh = csr_matrix(c_x).T

                # jacobian matrix H
                H_temp = sem.create_hx_jacobian(E)
                C_temp = sem.create_cx_jacobian(E, p_zero_injections, q_zero_injections)
                H, C = csr_matrix(H_temp), csr_matrix(C_temp)

                # gain matrix G_m
                # G_m = H^t * R^-1 * H
                G_m = H.T * (r_inv * H)

                # building a new gain matrix for new constraints.
                A_1 = vstack([G_m, C])
                c_ax = hstack([C, np.zeros((C.shape[0], C.shape[0]))])
                c_xT = c_ax.T
                M_tx = csr_matrix(hstack((A_1, c_xT)))  # again adding to the new gain matrix
                rhs = H.T * (r_inv * r)  # original right hand side
                C_rhs = vstack((rhs, -c_rxh))  # creating the righ hand side with new constraints

                # state vector difference d_E and update E
                d_E_ext = spsolve(M_tx, C_rhs)
                E_ext += d_E_ext.ravel()
                E = E_ext[:E.shape[0]]
                eppci.update_E(E)

                # prepare next iteration
                cur_it += 1
                current_error = np.max(np.abs(d_E_ext[:len(eppci.non_slack_buses) + num_bus]))
                self.logger.debug("Current error: {:.7f}".format(current_error))
            except np.linalg.linalg.LinAlgError:
                self.logger.error("A problem appeared while using the linear algebra methods."
                                  "Check and change the measurement set.")
                return False

        # check if the estimation is successfull
        self.check_result(current_error, cur_it)
        return eppci


class IRWLSAlgorithm(BaseAlgorithm):
    def estimate(self, eppci: ExtendedPPCI, estimator="wls", **kwargs):
        self.initialize(eppci)

        # matrix calculation object
        sem = get_estimator(BaseEstimatorIRWLS, estimator)(eppci, **kwargs)

        current_error, cur_it = 100., 0
        E = eppci.E
        while current_error > self.tolerance and cur_it < self.max_iterations:
            self.logger.debug("Starting iteration {:d}".format(1 + cur_it))
            try:
                # residual r
                r = csr_matrix(sem.create_rx(E)).T

                # jacobian matrix H
                H = csr_matrix(sem.create_hx_jacobian(E))

                # gain matrix G_m
                # G_m = H^t * Phi * H
                phi = csr_matrix(sem.create_phi(E))
                G_m = H.T * (phi * H)

                # state vector difference d_E and update E
                d_E = spsolve(G_m, H.T * (phi * r))
                E += d_E.ravel()
                eppci.update_E(E)

                # prepare next iteration
                cur_it += 1
                current_error = np.max(np.abs(d_E))
                self.logger.debug("Current error: {:.7f}".format(current_error))
            except np.linalg.linalg.LinAlgError:
                self.logger.error("A problem appeared while using the linear algebra methods."
                                  "Check and change the measurement set.")
                return False

        # check if the estimation is successfull
        self.check_result(current_error, cur_it)
        # update V/delta
        return eppci


class AFWLSAlgorithm(BaseAlgorithm):
    def __init__(self, tolerance, maximum_iterations, logger=std_logger):
        super(AFWLSAlgorithm, self).__init__(tolerance, maximum_iterations, logger)

        # Parameters for Bad data detection
        self.R_inv = None
        self.Gm = None
        self.r = None
        self.H = None
        self.hx = None
        self.iterations = None
        self.obj_func = None
        logging.basicConfig(level=logging.DEBUG)

    def estimate(self, eppci: ExtendedPPCI, **kwargs):
        self.initialize(eppci)
        # matrix calculation object
        sem = BaseAlgebra(eppci)

        current_error, cur_it = 100., 0
        # invert covariance matrix
        eppci.r_cov[eppci.r_cov<(10**(-5))] = 10**(-5)
        r_inv = csr_matrix(np.diagflat(1 / eppci.r_cov ** 2))
        E = eppci.E
        num_clusters = len(self.eppci["clusters"])
        while current_error > self.tolerance and cur_it < self.max_iterations:
            # self.logger.debug("Starting iteration {:d}".format(1 + cur_it))
            try:
                # residual r
                r = csr_matrix(sem.create_rx(E)).T

                # jacobian matrix H
                H = csr_matrix(sem.create_hx_jacobian(E))

                if cur_it == 0 and eppci.any_i_meas:
                    idx = eppci.idx_non_imeas
                    r_inv = r_inv[idx,:][:,idx]
                    r = r[idx,:]
                    H = H[idx,:]

                # gain matrix G_m
                G_m = H.T * (r_inv * H)
                norm_G = norm(G_m, np.inf)
                norm_invG = norm(inv(G_m), np.inf)
                cond = norm_G*norm_invG
                if cond > 10**18:
                    self.logger.warning("WARNING: Gain matrix is ill-conditioned: {:.2E}".format(cond))

                # state vector difference d_E
                d_E = spsolve(G_m, H.T * (r_inv * r))

                # Update E with d_E
                E += d_E.ravel()

                # log data 
                current_error = np.max(np.abs(d_E))
                # obj_func = (r.T*r_inv*r)[0,0]
                # self.logger.debug("Current delta_x: {:.7f}".format(current_error))
                # self.logger.debug("Current objective function value: {:.1f}".format(obj_func))

                # Restore full weighting matrix
                if cur_it == 0 and eppci.any_i_meas:
                    r_inv = csr_matrix(np.diagflat(1 / eppci.r_cov ** 2))

                # prepare next iteration
                cur_it += 1

            except np.linalg.linalg.LinAlgError:
                self.logger.error("A problem appeared while using the linear algebra methods."
                                  "Check and change the measurement set.")
                return False

        # check if the estimation is successfull
        self.check_result(current_error, cur_it)
        self.iterations = cur_it
        # self.obj_func = obj_func
        if self.successful:
            # store variables required for chi^2 and r_N_max test:
            self.R_inv = r_inv.toarray()
            self.Gm = G_m.toarray()
            self.r = r.toarray()
            self.H = H.toarray()
            # split voltage and allocation factor variables
            E1 = E[:-num_clusters]
            E2 = E[-num_clusters:]
            eppci.update_E(E1)
            eppci.clusters = E2
        return eppci


# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from time import perf_counter
import numpy as np

from pandapower.auxiliary import _sum_by_group, _check_if_numba_is_installed,\
    _check_bus_index_and_print_warning_if_high,\
    _check_gen_index_and_print_warning_if_high, \
    _add_pf_options, _add_ppc_options, _clean_up, sequence_to_phase, \
    phase_to_sequence, X012_to_X0, X012_to_X2, \
    I1_from_V012, S_from_VI_elementwise, V1_from_ppc, V_from_I,\
    combine_X012, I0_from_V012, I2_from_V012
from pandapower.powerflow import LoadflowNotConverged
from pandapower.build_bus import _add_ext_grid_sc_impedance
from pandapower.pypower.pfsoln import pfsoln
from pandapower.pypower.idx_gen import PG, QG, GEN_BUS
from pandapower.pypower.idx_bus import GS, BS, PD, QD, VM, VA
from pandapower.pd2ppc import _pd2ppc_recycle
from pandapower.pypower.makeYbus import makeYbus
from pandapower.pf.run_newton_raphson_pf import _run_newton_raphson_pf
from pandapower.pypower.bustypes import bustypes
from pandapower.run import _passed_runpp_parameters
from pandapower.results import _copy_results_ppci_to_ppc, _extract_results_3ph,\
    init_results
try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging
logger = logging.getLogger(__name__)


def _get_pf_variables_from_ppci(ppci):
    """
    Used for getting values for pfsoln function in one convinient function
    """
    # default arguments
    if ppci is None:
        ValueError('ppci is empty')
    # get data for calc
    base_mva, bus, gen, branch = \
        ppci["baseMVA"], ppci["bus"], ppci["gen"], ppci["branch"]
    # get bus index lists of each type of bus
    ref, pv, pq = bustypes(bus, gen)
    return base_mva, bus, gen, branch, ref, pv, pq


def _store_results_from_pf_in_ppci(ppci, bus, gen, branch):
    ppci["bus"], ppci["gen"], ppci["branch"] = bus, gen, branch
    return ppci


def _get_elements(params, net, element, phase, typ):
    sign = -1 if element.endswith("sgen") else 1
    elm = net[element].values
#   # Trying to find the column no for using numpy filters for active loads
    scaling = net[element].columns.get_loc("scaling")
    typ_col = net[element].columns.get_loc("type")  # Type = Delta or Wye load
    # active wye or active delta row selection
    active = (net["_is_elements"][element]) & (elm[:, typ_col] == typ)
    bus = [net[element].columns.get_loc("bus")]
    if len(elm):
        if element == 'load' or element == 'sgen':
            vl = elm[active, scaling].ravel()
            p_mw = [net[element].columns.get_loc("p_mw")]
            q_mvar = [net[element].columns.get_loc("q_mvar")]
            params['p'+phase+typ] = np.hstack([params['p'+phase+typ],
                                               elm[active, p_mw]/3 * vl * sign])
            params['q'+phase+typ] = np.hstack([params['q'+phase+typ],
                                               (elm[active, q_mvar]/3) * vl * sign])
            params['b'+typ] = np.hstack([params['b'+typ],
                                         elm[active, bus].astype(np.int64)])

        elif element.startswith('asymmetric'):
            vl = elm[active, scaling].ravel()
            p = {'a': net[element].columns.get_loc("p_a_mw"),
                 'b': net[element].columns.get_loc("p_b_mw"),
                 'c': net[element].columns.get_loc("p_c_mw")
                 }
            q = {'a': net[element].columns.get_loc("q_a_mvar"),
                 'b': net[element].columns.get_loc("q_b_mvar"),
                 'c': net[element].columns.get_loc("q_c_mvar")
                 }

            params['p'+phase+typ] = np.hstack([params['p'+phase+typ],
                                               elm[active, p[phase]] * vl * sign])
            params['q'+phase+typ] = np.hstack([params['q'+phase+typ],
                                               elm[active, q[phase]] * vl * sign])
            params['b'+typ] = np.hstack([params['b'+typ],
                                         elm[active, bus].astype(np.int64)])
    return params


def _load_mapping(net, ppci1):
    """
    Takes three phase P, Q values from PQ elements
    sums them up for each bus
    maps them in ppc bus order and forms s_abc matrix
    """
    bus_lookup = net["_pd2ppc_lookups"]["bus"]
    params = dict()
    phases = ['a', 'b', 'c']
    load_types = ['wye', 'delta']
    load_elements = ['load', 'asymmetric_load', 'sgen', 'asymmetric_sgen']
# =============================================================================
#        Loop to initialize and feed s_abc wye and delta values
# =============================================================================
    for phase in phases:
        for typ in load_types:
            params['S'+phase+typ] = (ppci1["bus"][:, PD] +
                                     ppci1["bus"][:, QD]*1j)*0
            params['p'+phase+typ] = np.array([])  # p values from loads/sgens
            params['q'+phase+typ] = np.array([])  # q values from loads/sgens
            params['P'+phase+typ] = np.array([])  # Aggregated Active Power
            params['Q'+phase+typ] = np.array([])  # Aggregated reactive Power
            params['b'+phase+typ] = np.array([], dtype=np.int64)  # bus map for phases
            params['b'+typ] = np.array([], dtype=np.int64)  # aggregated bus map(s_abc)
            for element in load_elements:
                _get_elements(params, net, element, phase, typ)
            # Mapping constant power loads to buses
            if params['b'+typ].size:
                params['b'+phase+typ] = bus_lookup[params['b'+typ]]
                params['b'+phase+typ], params['P'+phase+typ],\
                    params['Q'+phase+typ] = _sum_by_group(params['b'+phase+typ],
                                                          params['p'+phase+typ],
                                                          params['q'+phase+typ] * 1j)
                params['S'+phase+typ][params['b'+phase+typ]] = \
                    (params['P'+phase+typ] + params['Q'+phase+typ])
    Sabc_del = np.vstack((params['Sadelta'], params['Sbdelta'], params['Scdelta']))
    Sabc_wye = np.vstack((params['Sawye'], params['Sbwye'], params['Scwye']))
    # last return varaible left for constant impedance loads
    return Sabc_del, Sabc_wye


# =============================================================================
# 3 phase algorithm function
# =============================================================================
def runpp_3ph(net, calculate_voltage_angles=True, init="auto",
              max_iteration="auto", tolerance_mva=1e-8, trafo_model='t',
              trafo_loading="current", enforce_q_lims=False, numba=True,
              recycle=None, check_connectivity=True, switch_rx_ratio=2.0,
              delta_q=0, v_debug=False, **kwargs):
    """
 runpp_3ph: Performs Unbalanced/Asymmetric/Three Phase Load flow

    INPUT:
        **net** - The pandapower format network

    OPTIONAL:
        **algorithm** (str, "nr") - algorithm that is used to solve the power
        flow problem.

            The following algorithms are available:

                - "nr" Newton-Raphson (pypower implementation with numba accelerations)

                Used only for positive sequence network

                Zero and Negative sequence networks use Current Injection method

                Vnew = Y.inv * Ispecified ( from s_abc/v_abc old)

                Icalculated = Y * Vnew


        **calculate_voltage_angles** (bool, "auto") - consider voltage angles
        in loadflow calculation

            If True, voltage angles of ext_grids and transformer shifts are
            considered in the loadflow calculation. Considering the voltage
            angles is only necessary in meshed networks that are usually
            found in higher voltage levels. calculate_voltage_angles
            in "auto" mode defaults to:

                - True, if the network voltage level is above 70 kV
                - False otherwise

            The network voltage level is defined as the maximum rated voltage
            of any bus in the network that is connected to a line.


        **max_iteration** (int, "auto") - maximum number of iterations carried
        out in the power flow algorithm.

            In "auto" mode, the default value depends on the power flow solver:

                - 10 for "nr"

            For three phase calculations, its extended to 3 * max_iteration

        **tolerance_mva** (float, 1e-8) - loadflow termination condition
        referring to P / Q mismatch of node power in MVA

        **trafo_model**

            - transformer equivalent models
            - "t" - transformer is modeled as equivalent with the T-model.
            - "pi" - This is not recommended, since it is less exact than the T-model.
                So, for three phase load flow, its not implemented


        **trafo_loading** (str, "current") - mode of calculation for
        transformer loading

            Transformer loading can be calculated relative to the rated
            current or the rated power. In both cases the overall transformer
            loading is defined as the maximum loading on the two sides of
            the transformer.

            - "current"- transformer loading is given as ratio of current
                        flow and rated current of the transformer. This is the recommended
                        setting, since thermal as well as magnetic effects in the
                        transformer depend on the current.
            - "power" - transformer loading is given as ratio of apparent
                        power flow to the rated apparent power of the transformer.

        **enforce_q_lims** (bool, False) (Not tested with 3 Phase load flow)
        - respect generator reactive power limits

            If True, the reactive power limits in net.gen.max_q_mvar/min_q_mvar
            are respected in the loadflow. This is done by running a second
            loadflow if reactive power limits are violated at any generator,
            so that the runtime for the loadflow will increase if reactive
            power has to be curtailed.

            Note: enforce_q_lims only works if algorithm="nr"!


        **check_connectivity** (bool, True) - Perform an extra connectivity
        test after the conversion from pandapower to PYPOWER.

            If True, an extra connectivity test based on SciPy Compressed
            Sparse Graph Routines is perfomed. If check finds unsupplied buses,
            they are set out of service in the ppc

        **voltage_depend_loads** (bool, True)
        (Not tested with 3 Phase load flow) - consideration of voltage-dependent loads.

            If False, ``net.load.const_z_percent`` and ``net.load.const_i_percent``
            are not considered, i.e. ``net.load.p_mw`` and ``net.load.q_mvar``
            are considered as constant-power loads.

        **consider_line_temperature** (bool, False) (Not tested with 3 Phase
        load flow) - adjustment of line impedance based on provided line temperature.

            If True, ``net.line`` must contain a column ``temperature_degree_celsius``.
            The temperature dependency coefficient alpha must be provided in
            the ``net.line.alpha`` column, otherwise the default value of 0.004 is used.


        **KWARGS**:

        **numba** (bool, True) - Activation of numba JIT compiler in the
        newton solver

            If set to True, the numba JIT compiler is used to generate
            matrices for the powerflow, which leads to significant speed
            improvements.

        **switch_rx_ratio** (float, 2) (Not tested with 3 Phase load flow) - rx_ratio of bus-bus-switches. If the impedance of switches
        defined in net.switch.z_ohm is zero, buses connected by a closed bus-bus switch are fused to
        model an ideal bus. Closed bus-bus switches, whose impedance z_ohm is not zero, are modelled
        as branches with resistance and reactance according to net.switch.z_ohm and switch_rx_ratio.

        **delta_q**

        (Not tested with 3 Phase load flow) - Reactive power tolerance for option "enforce_q_lims"
        in kvar - helps convergence in some cases.

        **trafo3w_losses**

        (Not tested with 3 Phase load flow) - defines where open loop losses of three-winding
        transformers are considered. Valid options are "hv", "mv", "lv"
        for HV/MV/LV side or "star" for the star point.

        **v_debug** (bool, False) (Not tested with 3 Phase load flow) - if True,
        voltage values in each newton-raphson iteration are logged in the ppc.

        **init_vm_pu** (string/float/array/Series, None) (Not tested with 3
        Phase load flow) - Allows to define initialization specifically for
        voltage magnitudes. Only works with ``init == "auto"``!

            - "auto": all buses are initialized with the mean value of all
              voltage controlled elements in the grid
            - "flat" for flat start from 1.0
            - "results": voltage magnitude vector is taken from result table
            - a float with which all voltage magnitudes are initialized
            - an iterable with a voltage magnitude value for each bus
              (length and order has to match with the buses in net.bus)
            - a pandas Series with a voltage magnitude value for each bus
              (indexes have to match the indexes in net.bus)

         **init_va_degree** (string/float/array/Series, None) (Not tested with
         3 Phase load flow) - Allows to define initialization specifically for voltage angles.
         Only works with ``init == "auto"``!

            - "auto": voltage angles are initialized from DC power flow
              if angles are calculated or as 0 otherwise
            - "dc": voltage angles are initialized from DC power flow
            - "flat" for flat start from 0
            - "results": voltage angle vector is taken from result table
            - a float with which all voltage angles are initialized
            - an iterable with a voltage angle value for each bus (length
              and order has to match with the buses in net.bus)
            - a pandas Series with a voltage angle value for each bus (indexes
              have to match the indexes in net.bus)

        **recycle** (dict, none) - Reuse of internal powerflow variables for
        time series calculation.

            Contains a dict with the following parameters:
            bus_pq: If True PQ values of buses are updated

            gen: If True Sbus and the gen table in the ppc are recalculated

            Ybus: If True the admittance matrix (Ybus, Yf, Yt) is taken from

            ppc["internal"] and not reconstructed

        **neglect_open_switch_branches** (bool, False)

        (Not tested with 3 Phase load flow) - If True no auxiliary
        buses are created for branches when switches are opened at the branch.
        Instead branches are set out of service

    SEE ALSO:
         pp.add_zero_impedance_parameters(net):
         To add zero sequence parameters into network from the standard type

    EXAMPLES:
        Use this module like this:

        .. code-block:: python

            from pandapower.pf.runpp_3ph import runpp_3ph
            runpp_3ph(net)

    NOTES:
        - Three phase load flow uses Sequence Frame for power flow solution.
        - Three phase system is modelled with earth return.
        - PH-E load type is called as wye since Neutral and Earth are considered same
        - This solver has proved successful only for Earthed transformers (i.e Dyn,Yyn,YNyn & Yzn vector groups)
    """
    # =============================================================================
    # pandapower settings
    # =============================================================================
    overrule_options = {}
    if "user_pf_options" in net.keys() and len(net.user_pf_options) > 0:
        passed_parameters = _passed_runpp_parameters(locals())
        overrule_options = {key: val for key, val in net.user_pf_options.items()
                            if key not in passed_parameters.keys()}
    if numba:
        numba = _check_if_numba_is_installed()

    ac = True
    mode = "pf_3ph"  # TODO: Make valid modes (pf, pf_3ph, se, etc.) available in seperate file (similar to idx_bus.py)
#    v_debug = kwargs.get("v_debug", False)
    copy_constraints_to_ppc = False
    if trafo_model == 'pi':
        raise NotImplementedError("Three phase Power Flow doesnot support pi model because of lack of accuracy")
#    if calculate_voltage_angles == "auto":
#        calculate_voltage_angles = False
#        hv_buses = np.where(net.bus.vn_kv.values > 70)[0]  # Todo: Where does that number come from?
#        if len(hv_buses) > 0:
#            line_buses = net.line[["from_bus", "to_bus"]].values.flatten()
#            if len(set(net.bus.index[hv_buses]) & set(line_buses)) > 0:
    # scipy spsolve options in NR power flow
    use_umfpack = kwargs.get("use_umfpack", True)
    permc_spec = kwargs.get("permc_spec", None)
    calculate_voltage_angles = True
    if init == "results" and net.res_bus_3ph.empty:
        init = "auto"
    if init == "auto":
        init = "dc" if calculate_voltage_angles else "flat"
    default_max_iteration = {"nr": 30, "bfsw": 10, "gs": 10000, "fdxb": 30,
                             "fdbx": 30}
    if max_iteration == "auto":
        max_iteration = default_max_iteration["nr"]

    neglect_open_switch_branches = kwargs.get("neglect_open_switch_branches", False)
    only_v_results = kwargs.get("only_v_results", False)
    net._options = {}
    _add_ppc_options(net, calculate_voltage_angles=calculate_voltage_angles,
                     trafo_model=trafo_model, check_connectivity=check_connectivity,
                     mode=mode, switch_rx_ratio=switch_rx_ratio,
                     init_vm_pu=init, init_va_degree=init,
                     enforce_q_lims=enforce_q_lims, recycle=None,
                     voltage_depend_loads=False, delta=delta_q,
                     neglect_open_switch_branches=neglect_open_switch_branches
                     )
    _add_pf_options(net, tolerance_mva=tolerance_mva, trafo_loading=trafo_loading,
                    numba=numba, ac=ac, algorithm="nr", max_iteration=max_iteration,
                    only_v_results=only_v_results, v_debug=v_debug, use_umfpack=use_umfpack,
                    permc_spec=permc_spec, lightsim2grid=False)
    net._options.update(overrule_options)
    _check_bus_index_and_print_warning_if_high(net)
    _check_gen_index_and_print_warning_if_high(net)
    # =========================================================================
    # pd2ppc conversion
    # =========================================================================
    _, ppci1 = _pd2ppc_recycle(net, 1, recycle=recycle)

    _, ppci2 = _pd2ppc_recycle(net, 2, recycle=recycle)
    gs_eg, bs_eg = _add_ext_grid_sc_impedance(net, ppci2)

    _, ppci0 = _pd2ppc_recycle(net, 0, recycle=recycle)

    _,        bus0, gen0, branch0,      _,      _,      _ = _get_pf_variables_from_ppci(ppci0)
    base_mva, bus1, gen1, branch1, sl_bus,      _, pq_bus = _get_pf_variables_from_ppci(ppci1)
    _,        bus2, gen2, branch2,      _,      _,      _ = _get_pf_variables_from_ppci(ppci2)

    # initialize the results after the conversion to ppc is done, otherwise init=results does not work
    init_results(net, "pf_3ph")

# =============================================================================
#     P Q values aggragated and summed up for each bus to make s_abc matrix
#     s_abc for wye connections ; s_abc_delta for delta connection
# =============================================================================
    s_abc_delta, s_abc = _load_mapping(net, ppci1)
    # =========================================================================
    # Construct Sequence Frame Bus admittance matrices Ybus
    # =========================================================================

    ppci0, ppci1, ppci2, y_0_pu, y_1_pu, y_2_pu, y_0_f, y_1_f, _,\
        y_0_t, y_1_t, _ = _get_y_bus(ppci0, ppci1, ppci2, recycle)
    # =========================================================================
    # Initial voltage values
    # =========================================================================
    nb = ppci1["bus"].shape[0]

    # make sure flat start is always respected, even with other voltage data in recycled ppc
    if init == "flat":
        v_012_it = np.zeros((3, nb), dtype=np.complex128)
        v_012_it[1, :] = 1.0
    else:
        v_012_it = np.concatenate(
            [np.array(ppc["bus"][:, VM] * np.exp(1j * np.deg2rad(ppc["bus"][:, VA]))).reshape((1, nb))
             for ppc in (ppci0, ppci1, ppci2)], axis=0).astype(np.complex128)

    # For Delta transformation:
    # Voltage changed from line-earth to line-line using V_T
    # s_abc/v_abc will now give line-line currents. This is converted to line-earth
    # current using I-T
    v_del_xfmn = np.array([[1, -1, 0],
                           [0, 1, -1],
                           [-1, 0, 1]])
    i_del_xfmn = np.array([[1, 0, -1],
                           [-1, 1, 0],
                           [0, -1, 1]])
    v_abc_it = sequence_to_phase(v_012_it)

    # =========================================================================
    #             Iteration using Power mismatch criterion
    # =========================================================================
    outer_tolerance_mva = 3e-8
    count = 0
    s_mismatch = np.array([[True], [True]], dtype=bool)
    t0 = perf_counter()
    while (s_mismatch > outer_tolerance_mva).any() and count < max_iteration:
        # =====================================================================
        #     Voltages and Current transformation for PQ and Slack bus
        # =====================================================================
        s_abc_pu = -np.divide(s_abc, ppci1["baseMVA"])
        s_abc_delta_pu = -np.divide(s_abc_delta, ppci1["baseMVA"])

        i_abc_it_wye = (np.divide(s_abc_pu, v_abc_it)).conjugate()
        i_abc_it_delta = np.matmul(i_del_xfmn, (np.divide(s_abc_delta_pu, np.matmul
                                                          (v_del_xfmn, v_abc_it))).conjugate())

        # For buses with both delta and wye loads we need to sum of their currents
        # to sum up the currents
        i_abc_it = i_abc_it_wye + i_abc_it_delta
        i012_it = phase_to_sequence(i_abc_it)
        v1_for_s1 = v_012_it[1, :]
        i1_for_s1 = -i012_it[1, :]
        v0_pu_it = X012_to_X0(v_012_it)
        v2_pu_it = X012_to_X2(v_012_it)
        i0_pu_it = X012_to_X0(i012_it)
        i2_pu_it = X012_to_X2(i012_it)
        s1 = np.multiply(v1_for_s1, i1_for_s1.conjugate())
        # =============================================================================
        # Current used to find S1 Positive sequence power
        # =============================================================================

        ppci1["bus"][pq_bus, PD] = np.real(s1[pq_bus]) * ppci1["baseMVA"]
        ppci1["bus"][pq_bus, QD] = np.imag(s1[pq_bus]) * ppci1["baseMVA"]
        # =============================================================================
        # Conduct Positive sequence power flow
        # =============================================================================
        _run_newton_raphson_pf(ppci1, net._options)
        # =============================================================================
        # Conduct Negative and Zero sequence power flow
        # =============================================================================
        v0_pu_it = V_from_I(y_0_pu, i0_pu_it)
        v2_pu_it = V_from_I(y_2_pu, i2_pu_it)
        # =============================================================================
        #    Evaluate Positive Sequence Power Mismatch
        # =============================================================================
        i1_from_v_it = I1_from_V012(v_012_it, y_1_pu).flatten()
        s_from_voltage = S_from_VI_elementwise(v1_for_s1, i1_from_v_it)
        v1_pu_it = V1_from_ppc(ppci1)

        v_012_new = combine_X012(v0_pu_it, v1_pu_it, v2_pu_it)

        s_mismatch = np.abs(np.abs(s1[pq_bus]) - np.abs(s_from_voltage[pq_bus]))
        v_012_it = v_012_new
        v_abc_it = sequence_to_phase(v_012_it)
        count += 1
    et = perf_counter() - t0
    success = (count < max_iteration)
    for ppc in [ppci0, ppci1, ppci2]:
        ppc["et"] = et
        ppc["success"] = success
    # TODO: Add reference to paper to explain the following steps
    # This is required since the ext_grid power results are not correct if its
    # not done
    ref, pv, pq = bustypes(ppci0["bus"], ppci0["gen"])
    ref_gens = ppci0["internal"]["ref_gens"]
    ppci0["bus"][ref, GS] -= gs_eg
    ppci0["bus"][ref, BS] -= bs_eg
    y_0_pu, y_0_f, y_0_t = makeYbus(ppci0["baseMVA"], ppci0["bus"], ppci0["branch"])
    # revert the change, otherwise repeated calculation with recycled elements will fail
    ppci0["bus"][ref, GS] += gs_eg
    ppci0["bus"][ref, BS] += bs_eg
    # Bus, Branch, and Gen  power values
    svc = ppci0["svc"]  # placeholder
    tcsc = ppci0["tcsc"]  # placeholder
    ssc = ppci0["ssc"]  # placeholder
    vsc = ppci0["vsc"]  # placeholder
    bus0, gen0, branch0 = pfsoln(base_mva, bus0, gen0, branch0, svc, tcsc, ssc, vsc, y_0_pu, y_0_f, y_0_t,
                                 v_012_it[0, :].flatten(), sl_bus, ref_gens)
    bus1, gen1, branch1 = pfsoln(base_mva, bus1, gen1, branch1, svc, tcsc, ssc, vsc, y_1_pu, y_1_f, y_1_t,
                                 v_012_it[1, :].flatten(), sl_bus, ref_gens)
    bus2, gen2, branch2 = pfsoln(base_mva, bus2, gen2, branch2, svc, tcsc, ssc, vsc, y_1_pu, y_1_f, y_1_t,
                                 v_012_it[2, :].flatten(), sl_bus, ref_gens)
    ppci0 = _store_results_from_pf_in_ppci(ppci0, bus0, gen0, branch0)
    ppci1 = _store_results_from_pf_in_ppci(ppci1, bus1, gen1, branch1)
    ppci2 = _store_results_from_pf_in_ppci(ppci2, bus2, gen2, branch2)
    i_012_res = _current_from_voltage_results(y_0_pu, y_1_pu, v_012_it)
    s_012_res = S_from_VI_elementwise(v_012_it, i_012_res) * ppci1["baseMVA"]
    eg_is_mask = net["_is_elements"]['ext_grid']
    ext_grid_lookup = net["_pd2ppc_lookups"]["ext_grid"]
    eg_is_idx = net["ext_grid"].index.values[eg_is_mask]
    eg_idx_ppc = ext_grid_lookup[eg_is_idx]
    """ # 2 ext_grids Fix: Instead of the generator index, bus indices of the generators are used"""
    eg_bus_idx_ppc = np.real(ppci1["gen"][eg_idx_ppc, GEN_BUS]).astype(np.int64)

    ppci0["gen"][eg_idx_ppc, PG] = s_012_res[0, eg_bus_idx_ppc].real
    ppci1["gen"][eg_idx_ppc, PG] = s_012_res[1, eg_bus_idx_ppc].real
    ppci2["gen"][eg_idx_ppc, PG] = s_012_res[2, eg_bus_idx_ppc].real
    ppci0["gen"][eg_idx_ppc, QG] = s_012_res[0, eg_bus_idx_ppc].imag
    ppci1["gen"][eg_idx_ppc, QG] = s_012_res[1, eg_bus_idx_ppc].imag
    ppci2["gen"][eg_idx_ppc, QG] = s_012_res[2, eg_bus_idx_ppc].imag

    ppc0 = net["_ppc0"]
    ppc1 = net["_ppc1"]
    ppc2 = net["_ppc2"]

    # ppci doesn't contain out of service elements, but ppc does -> copy results accordingly
    ppc0 = _copy_results_ppci_to_ppc(ppci0, ppc0, mode=mode)
    ppc1 = _copy_results_ppci_to_ppc(ppci1, ppc1, mode=mode)
    ppc2 = _copy_results_ppci_to_ppc(ppci2, ppc2, mode=mode)

    _extract_results_3ph(net, ppc0, ppc1, ppc2)

    #    Raise error if PF was not successful. If DC -> success is always 1

    if not ppci0["success"]:
        net["converged"] = False
        _clean_up(net, res=False)
        raise LoadflowNotConverged("Power Flow {0} did not converge after {1} iterations!".format("nr", count))
    else:
        net["converged"] = True

    _clean_up(net)


def _current_from_voltage_results(y_0_pu, y_1_pu, v_012_pu):
    I012_pu = combine_X012(I0_from_V012(v_012_pu, y_0_pu),
                           I1_from_V012(v_012_pu, y_1_pu),
                           I2_from_V012(v_012_pu, y_1_pu))
    return I012_pu


def _get_y_bus(ppci0, ppci1, ppci2, recycle):
    if recycle and recycle["Ybus"] and ppci0["internal"]["Ybus"].size and \
            ppci1["internal"]["Ybus"].size and ppci2["internal"]["Ybus"].size:
        y_0_bus, y_0_f, y_0_t = ppci0["internal"]['Ybus'], ppci0["internal"]['Yf'], ppci0["internal"]['Yt']
        y_1_bus, y_1_f, y_1_t = ppci1["internal"]['Ybus'], ppci1["internal"]['Yf'], ppci1["internal"]['Yt']
        y_2_bus, y_2_f, y_2_t = ppci2["internal"]['Ybus'], ppci2["internal"]['Yf'], ppci2["internal"]['Yt']
    else:
        # build admittance matrices
        y_0_bus, y_0_f, y_0_t = makeYbus(ppci0["baseMVA"], ppci0["bus"], ppci0["branch"])
        y_1_bus, y_1_f, y_1_t = makeYbus(ppci1["baseMVA"], ppci1["bus"], ppci1["branch"])
        y_2_bus, y_2_f, y_2_t = makeYbus(ppci2["baseMVA"], ppci2["bus"], ppci2["branch"])
        if recycle and recycle["Ybus"]:
            ppci0["internal"]['Ybus'], ppci0["internal"]['Yf'], ppci0["internal"]['Yt'] = y_0_bus, y_0_f, y_0_t
            ppci1["internal"]['Ybus'], ppci1["internal"]['Yf'], ppci1["internal"]['Yt'] = y_1_bus, y_1_f, y_1_t
            ppci2["internal"]['Ybus'], ppci2["internal"]['Yf'], ppci2["internal"]['Yt'] = y_2_bus, y_2_f, y_2_t

    return ppci0, ppci1, ppci2, y_0_bus, y_1_bus, y_2_bus, y_0_f, y_1_f, y_2_f, y_0_t, y_1_t, y_2_t




# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from pandapower.control.controller.characteristic_control import CharacteristicControl
from pandapower.control.util.characteristic import Characteristic


class VmSetTapControl(CharacteristicControl):
    """
    Controller that adjusts the setpoint of a local tap changer voltage control based on a load flow result (e.g. p_lv_mw, i_lv_ka etc.)
    according to a defined characteristic.

    INPUT:
        **net** (attrdict) - Pandapower net

        **cid** (int) - ID of the tap changer controller, an attribute of which is controlled

        **variable** (float) - Variable from the result table that is used for the characteristic

    OPTIONAL:

        **in_service** (bool, True) - Indicates if the controller is currently in_service

        **drop_same_existing_ctrl** (bool, False) - Indicates if already existing controllers of the same type and with the same matching parameters (e.g. at same element) should be dropped
    """

    def __init__(self, net, controller_index, characteristic_index, variable='p_hv_mw', tol=1e-3, in_service=True,
                 order=0, level=0, drop_same_existing_ctrl=False, matching_params=None, **kwargs):
        if matching_params is None:
            matching_params = {"cid": controller_index, 'variable': variable}
        c = net.controller.at[controller_index, 'object']
        super().__init__(
            net, output_element="controller", output_variable="object.vm_set_pu",
            output_element_index=controller_index, input_element="res_" + c.element,
            input_variable=variable, input_element_index=c.element_index,
            characteristic_index=characteristic_index, tol=tol, in_service=in_service, order=order,
            level=level, drop_same_existing_ctrl=drop_same_existing_ctrl,
            matching_params=matching_params, **kwargs)



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import os

from pandapower.__init__ import pp_dir
from pandapower.file_io import from_json
from pandapower.run import runpp
from pandapower.toolbox.grid_modification import drop_elements, select_subnet
from pandapower.topology.create_graph import create_nxgraph
from pandapower.topology.graph_searches import connected_components


def lv_schutterwald(separation_by_sub=False, include_heat_pumps=False, **kwargs):
    """
    Loads the Schutterwald network, a generic 0.4 kV network serviced by 14 MV/LV transformer
    stations of the Oberrhein network.
    The network supplies 1506 customers with the option of including 1251 heat pumps.

    The network also includes geographical information of lines and buses for plotting.

    Source: https://doi.org/10.3390/en13164052

    OPTIONAL:
        **separation_by_sub** - (bool, False): if True, the network gets separated into 14
        sections, referring to their substations

        **include_heat_pumps** - (bool, False): if True, the heat pumps from the study are
        included in the network

    OUTPUT:
         **net** - pandapower network

         **subnets** (list) - all sections of the pandapower network

    EXAMPLE:

        >>> from pandapower.networks import lv_schutterwald
        >>> net = lv_schutterwald()

        or with separation

        >>> net_list = lv_schutterwald(separation_by_sub=True)
    """

    net = from_json(os.path.join(pp_dir, "networks", "lv_schutterwald.json"), **kwargs)

    # modifications on the original file
    # geo.convert_crs(net, epsg_out=4326, epsg_in=31467)
    # geo.convert_geodata_to_geojson(net, lonlat=False)
    # drop_inactive_elements(net)
    # net.load.replace({"type": "WP"}, "HP", inplace=True)
    # net.load.replace({"type": "HA"}, "HH", inplace=True)
    # net.load.name = net.load.name.str.replace("H", "HH", regex=False)
    # net.load.name = net.load.name.str.replace("WP", "HP", regex=False)

    if not include_heat_pumps:
        drop_elements(net, "load", net.load.loc[net.load.type == "HP"].index)

    subnets = list()
    if separation_by_sub:
        # creating multigraph
        mg = create_nxgraph(net)
        # clustering connected buses
        zones = [list(area) for area in connected_components(mg)]
        for i, zone in enumerate(zones):
            net1 = select_subnet(net, buses=zone, include_switch_buses=False,
                                 include_results=True, keep_everything_else=True)
            runpp(net1)
            net1.name = f'LV Schutterwald {i}'
            subnets.append(net1)
        return subnets

    runpp(net)
    net.name = 'LV Schutterwald'
    return net



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np

from pandapower.auxiliary import ensure_iterability
from pandapower.toolbox.element_selection import pp_elements

try:
    from networkx.utils.misc import graphs_equal
    GRAPHS_EQUAL_POSSIBLE = True
except ImportError:
    GRAPHS_EQUAL_POSSIBLE = False

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def signing_system_value(element_type):
    """
    Returns a 1 for all bus elements using the consumver viewpoint and a -1 for all bus elements
    using the generator viewpoint.
    """
    generator_viewpoint_ets = ["ext_grid", "gen", "sgen"]
    if element_type in generator_viewpoint_ets:
        return -1
    elif element_type in pp_elements(bus=False, other_elements=False):
        return 1
    else:
        raise ValueError("This function is defined for bus and branch elements, not for "
                         f"'{element_type}'.")


def pq_from_cosphi(s, cosphi, qmode, pmode):
    """
    Calculates P/Q values from rated apparent power and cosine(phi) values.

       - s: rated apparent power
       - cosphi: cosine phi of the
       - qmode: "underexcited" (Q absorption, decreases voltage) or "overexcited" (Q injection, increases voltage)
       - pmode: "load" for load or "gen" for generation

    As all other pandapower functions this function is based on the consumer viewpoint. For active
    power, that means that loads are positive and generation is negative. For reactive power,
    underexcited behavior (Q absorption, decreases voltage) is modeled with positive values,
    overexcited behavior (Q injection, increases voltage) with negative values.
    """
    if hasattr(s, "__iter__"):
        len_ = len(s)
    elif hasattr(cosphi, "__iter__"):
        len_ = len(cosphi)
    elif not isinstance(qmode, str) and hasattr(qmode, "__iter__"):
        len_ = len(qmode)
    elif not isinstance(pmode, str) and hasattr(pmode, "__iter__"):
        len_ = len(pmode)
    else:
        return _pq_from_cosphi(s, cosphi, qmode, pmode)
    return _pq_from_cosphi_bulk(s, cosphi, qmode, pmode, len_=len_)


def _pq_from_cosphi(s, cosphi, qmode, pmode):
    if qmode in ("ind", "cap"):
        logger.warning('capacitive or inductive behavior will be replaced by more clear terms ' +
                       '"underexcited" (Q absorption, decreases voltage) and "overexcited" ' +
                       '(Q injection, increases voltage). Please use "underexcited" ' +
                       'in place of "ind" and "overexcited" in place of "cap".')
    if qmode == "ind" or qmode == "underexcited":
        qsign = 1
    elif qmode == "cap" or qmode == "overexcited":
        qsign = -1
    else:
        raise ValueError('Unknown mode %s - specify "underexcited" (Q absorption, decreases voltage'
                         ') or "overexcited" (Q injection, increases voltage)' % qmode)

    if pmode == "load":
        psign = 1
    elif pmode == "gen":
        psign = -1
    else:
        raise ValueError('Unknown mode %s - specify "load" or "gen"' % pmode)

    p = s * cosphi
    q = psign * qsign * np.sqrt(s ** 2 - p ** 2)
    return p, q


def _pq_from_cosphi_bulk(s, cosphi, qmode, pmode, len_=None):
    if len_ is None:
        s = np.array(ensure_iterability(s))
        len_ = len(s)
    else:
        s = np.array(ensure_iterability(s, len_))
    cosphi = np.array(ensure_iterability(cosphi, len_))
    qmode = np.array(ensure_iterability(qmode, len_))
    pmode = np.array(ensure_iterability(pmode, len_))

    # "ind" -> "underexcited", "cap" -> "overexcited"
    is_ind = qmode == "ind"
    is_cap = qmode == "cap"
    if any(is_ind) or any(is_cap):
        logger.warning('capacitive or inductive behavior will be replaced by more clear terms ' +
                       '"underexcited" (Q absorption, decreases voltage) and "overexcited" ' +
                       '(Q injection, increases voltage). Please use "underexcited" ' +
                       'in place of "ind" and "overexcited" in place of "cap".')
    qmode[is_ind] = "underexcited"
    qmode[is_cap] = "overexcited"

    # qmode consideration
    unknown_qmode = set(qmode) - set(["underexcited", "overexcited"])
    if len(unknown_qmode):
        raise ValueError("Unknown qmodes: " + str(list(unknown_qmode)))
    qsign = np.ones(qmode.shape)
    qsign[qmode == "overexcited"] = -1

    # pmode consideration
    unknown_pmode = set(pmode) - set(["load", "gen"])
    if len(unknown_pmode):
        raise ValueError("Unknown pmodes: " + str(list(unknown_pmode)))
    psign = np.ones(pmode.shape)
    psign[pmode == "gen"] = -1

    # calculate p and q
    p = s * cosphi
    q = psign * qsign * np.sqrt(s ** 2 - p ** 2)

    return p, q


def cosphi_from_pq(p, q):
    """
    Analog to pq_from_cosphi, but the other way around.
    In consumer viewpoint (pandapower): "underexcited" (Q absorption, decreases voltage) and
    "overexcited" (Q injection, increases voltage)
    """
    if hasattr(p, "__iter__"):
        len_ = len(p)
    elif hasattr(q, "__iter__"):
        len_ = len(q)
    else:
        return _cosphi_from_pq(p, q)
    return _cosphi_from_pq_bulk(p, q, len_=len_)


def _cosphi_from_pq(p, q):
    if p == 0:
        cosphi = np.nan
        logger.warning("A cosphi from p=0 is undefined.")
    else:
        cosphi = np.cos(np.arctan(q / p))
    s = (p ** 2 + q ** 2) ** 0.5
    pmode = ["undef", "load", "gen"][int(np.sign(p))]
    qmode = ["underexcited", "underexcited", "overexcited"][int(np.sign(q))]
    return cosphi, s, qmode, pmode


def _cosphi_from_pq_bulk(p, q, len_=None):
    if len_ is None:
        p = np.array(ensure_iterability(p))
        len_ = len(p)
    else:
        p = np.array(ensure_iterability(p, len_))
    q = np.array(ensure_iterability(q, len_))
    p_is_zero = np.array(p == 0)
    cosphi = np.empty(p.shape)
    if sum(p_is_zero):
        cosphi[p_is_zero] = np.nan
        logger.warning("A cosphi from p=0 is undefined.")
    cosphi[~p_is_zero] = np.cos(np.arctan(q[~p_is_zero] / p[~p_is_zero]))
    s = (p ** 2 + q ** 2) ** 0.5
    pmode = np.array(["undef", "load", "gen"])[np.sign(p).astype(np.int64)]
    qmode = np.array(["underexcited", "underexcited", "overexcited"])[np.sign(q).astype(np.int64)]
    return cosphi, s, qmode, pmode

# -------------------------------------------------------------------------------------------------
"""
Functions with numeric encoding of qmodes "underexcited" or "overexcited":
pos_neg: cosphi is positive for positive q values and negative otherwise
pos: cosphi is not in [0, 1] and [-1, -0] but in [0, 1] and [1, 2]
"""
# -------------------------------------------------------------------------------------------------


def cosphi_pos_neg_from_pq(p, q):
    """Returns the cosphi value(s) for given active and reactive power(s).
    Positive q values lead to positive cosphi values, negative to negative.
    If p, q come from generator view point, positive cosphi correspond to overexcited behaviour
    while in load view point positive cosphi correspond to underexcited behaviour.

    Parameters
    ----------
    p : float(s)
        active power value(s)
    q : float(s)
        reactive power values

    Examples
    --------
    >>> import numpy as np
    >>> from pandapower.toolbox import cosphi_pos_neg_from_pq
    >>> np.round(cosphi_pos_neg_from_pq(0.76, 0.25), 5)
    0.94993
    >>> np.round(cosphi_pos_neg_from_pq(0.76, -0.25), 5)
    -0.94993
    >>> np.round(cosphi_pos_neg_from_pq([0.76, 0.76, 0.76, 0.76], [0.25, -0.25, 0, 0.1]), 5)
    array([ 0.94993,  -0.94993,  1.     ,  0.99145])
    >>> np.round(cosphi_pos_neg_from_pq([0.76, 0.76, -0.76, 0.76, 0, 0.1],
    ...                                 [0.25, -0.25, 0.25, 0.1, 0.1, 0]), 5)
    array([ 0.94993,  -0.94993,  0.94993, 0.99145, nan, 1])

    Returns
    -------
    [float, np.array]
        cosphi values where

    See also
    --------
    pandapower.toolbox.cosphi_from_pq
    pandapower.toolbox.cosphi_to_pos
    """
    cosphis = cosphi_from_pq(p, q)[0]
    return np.copysign(cosphis, q)



def cosphi_to_pos(cosphi):
    """ Signed cosphi values are converted into positive cosphi values from 0 to 1 and 1 to 2.

    Examples
    --------
    >>> cosphi_to_pos(0.96)
    0.96
    >>> cosphi_to_pos(-0.94)
    1.06
    >>> cosphi_to_pos(-0.96)
    1.04
    >>> cosphi_to_pos([0.96, -0.94, -0.96])
    np.array([0.96, 1.06, 1.04])

    See also
    --------
    pandapower.toolbox.cosphi_from_pos
    pandapower.toolbox.cosphi_pos_neg_from_pq
    """
    multiple = not (isinstance(cosphi, float) or isinstance(cosphi, int))
    cosphi = np.array(ensure_iterability(cosphi))
    cosphi[cosphi < 0] += 2
    if multiple:
        return cosphi
    else:
        return cosphi[0]


def cosphi_from_pos(cosphi):
    """ All positive cosphi values are converted back to signed cosphi values

    Examples
    --------
    >>> cosphi_from_pos(0.96)
    0.96
    >>> cosphi_from_pos(1.06)
    -0.94
    >>> cosphi_from_pos(1.04)
    -0.96
    >>> cosphi_from_pos([0.96, 1.06, 1.04])
    np.array([0.96, -0.94, -0.96])

    See also
    --------
    pandapower.toolbox.cosphi_to_pos
    pandapower.toolbox.cosphi_pos_neg_from_pq
    """
    multiple = not (isinstance(cosphi, float) or isinstance(cosphi, int))
    cosphi = np.array(ensure_iterability(cosphi))
    cosphi[cosphi > 1] -= 2
    if multiple:
        return cosphi
    else:
        return cosphi[0]



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from numpy import allclose, delete, any

from pandapower.auxiliary import _add_ppc_options
from pandapower.opf.validate_opf_input import _check_necessary_opf_parameters
from pandapower.powerflow import _pd2ppc
from pandapower.pypower.idx_brch import BR_G, BR_G_ASYM, BR_B_ASYM, BR_R_ASYM, BR_X_ASYM

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def to_ppc(net, calculate_voltage_angles=True, trafo_model="t", switch_rx_ratio=2,
           check_connectivity=True, voltage_depend_loads=False, init="results", mode=None,
           take_slack_vm_limits=True):
    """
    This function converts a pandapower net to a pypower case file.

    INPUT:
        **net** - The pandapower net.

    OPTIONAL:
        **calculate_voltage_angles** (bool, False) - consider voltage angles in loadflow calculation

        If True, voltage angles of ext_grids and transformer shifts are considered in the
        loadflow calculation. Considering the voltage angles is only necessary in meshed
        networks that are usually found in higher networks.

        **trafo_model** (str, "t") - transformer equivalent circuit model
        pandapower provides two equivalent circuit models for the transformer:

            - "t" - transformer is modeled as equivalent with the T-model.
            - "pi" - transformer is modeled as equivalent PI-model.

            The "pi" - Model is not recommended, since it is less exact than the T-model. It is only
            recommended for validation with other software that uses the pi-model.


        **switch_rx_ratio** (float, 2) - rx_ratio of bus-bus-switches. If the impedance of switches
        defined in net.switch.z_ohm is zero, buses connected by a closed bus-bus switch are fused to
        model an ideal bus. Closed bus-bus switches, whose impedance z_ohm is not zero, are modelled
        as branches with resistance and reactance according to net.switch.z_ohm and switch_rx_ratio.

        **check_connectivity** (bool, True) - Perform an extra connectivity test after the
        conversion from pandapower to PYPOWER

            If True, an extra connectivity test based on SciPy Compressed Sparse Graph Routines is
            perfomed. If the check finds unsupplied buses, they are set out of service in the ppc

        **voltage_depend_loads** (bool, False) - consideration of voltage-dependent loads.
        If False, net.load.const_z_percent and net.load.const_i_percent are not considered, i.e.
        net.load.p_mw and net.load.q_mvar are considered as constant-power loads.

        **init** (str, "results") - initialization method of the converter
        pandapower ppc converter supports two methods for initializing the converter:

            - "flat" - flat start with voltage of 1.0pu and angle of 0° at all PQ-buses and 0° for PV buses as initial solution

            - "results" - voltage vector from net.res_bus is used as initial solution.

        **mode** (str, None) - mode of power flow calculation type ("pf" - power flow, "opf" -
        optimal power flow or "sc" - short circuit). "mode" influences for instance whether opf
        cost data will be converted or which slack bus voltage limits are respected. If "mode"
        is None, cost data will be respected via mode="opf" if cost data exist.

        **take_slack_vm_limits** (bool, True) - Per default the voltage magnitude limits are assumed
        as setpoint of the slack unit (usually net.ext_grid.vm_pu). To replace that by values from
        net.bus[["min_vm_pu", "max_vm_pu"]], take_slack_vm_limits can be set to False.

    OUTPUT:
        **ppc** - The Pypower casefile for usage with pypower

    EXAMPLE:
        >>> from pandapower.converter import to_ppc
        >>> from pandapower.networks.power_system_test_cases import case9
        >>>
        >>> net = case9()
        >>> ppc = to_ppc(net)

    """
    if (not (net["poly_cost"].empty and net["pwl_cost"].empty) and
        mode is None) or mode == "opf":
        mode = "opf"
        _check_necessary_opf_parameters(net, logger)
    else:
        mode = "pf"

    # check init values
    if init != "flat":
        if not net.res_bus.shape[0]:
            raise UserWarning("res_bus is empty. Change the input parameter 'init' to 'flat' or "
                              "add result values to allow initialization with 'results'.")
        elif len(net.bus.index.difference(net.res_bus.index)):
            raise ValueError("The res_bus indices doesn't fit to the bus indices. Change the "
                             "input parameter 'init' to flat or correct the res_bus table.")

    # select elements in service
    net["_options"] = {}
    _add_ppc_options(net, calculate_voltage_angles=calculate_voltage_angles,
                     trafo_model=trafo_model, check_connectivity=check_connectivity,
                     mode=mode, switch_rx_ratio=switch_rx_ratio, init_vm_pu=init,
                     init_va_degree=init, enforce_q_lims=True,
                     recycle=None, voltage_depend_loads=voltage_depend_loads)

    if net["_options"]["voltage_depend_loads"] and not (
            allclose(net.load.const_z_percent.values, 0) and
            allclose(net.load.const_i_percent.values, 0)):
        logger.error("to_ppc() does not consider voltage depend loads. The z and i parts of "
                     "voltage depend loads are set to additional columns 13 and 14 but the p/q part"
                     " is still unchanged.")

    #  do the conversion
    _, ppci = _pd2ppc(net)

    # delete BR_G column as this is added to pandapower afterward, keeping it as extra variable
    # only keep the additional variables in ppc if any values are non-zero
    # todo: test saving & retrieving the data columns with to_ppc, from_ppc
    for ppc_name, brch_col in zip(['branch_r_asym', 'branch_x_asym', 'branch_g', 'branch_g_asym', 'branch_b_asym'],
                                  [BR_R_ASYM, BR_X_ASYM, BR_G, BR_G_ASYM, BR_B_ASYM]):
        if any(col := net._ppc["branch"][:, brch_col]):
            ppci[ppc_name] = col
    ppci['branch'] = delete(ppci['branch'], [BR_R_ASYM, BR_G_ASYM, BR_G, BR_G_ASYM, BR_B_ASYM], axis=1)
    if not take_slack_vm_limits:
        slack_bus = min(net.ext_grid.bus.loc[net.ext_grid.in_service].tolist() + \
                        net.gen.bus.loc[net.gen.slack & net.gen.in_service].tolist())
        slack_bus_ppci_pos = net.bus.index[net.bus.in_service].get_loc(slack_bus)
        ppci["bus"][slack_bus_ppci_pos, 11] = net.bus.max_vm_pu.at[slack_bus]
        ppci["bus"][slack_bus_ppci_pos, 12] = net.bus.min_vm_pu.at[slack_bus]

    return ppci



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import copy
from itertools import chain

import numpy as np
import pandas as pd
from pandapower.opf.validate_opf_input import _check_necessary_opf_parameters
from pandapower.toolbox import pp_elements

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def lf_info(net, numv=1, numi=2):  # pragma: no cover
    """
    Prints some basic information of the results in a net
    (max/min voltage, max trafo load, max line load).

    OPTIONAL:

        **numv** (integer, 1) - maximal number of printed maximal respectively minimal voltages

        **numi** (integer, 2) - maximal number of printed maximal loading at trafos or lines
    """
    logger.info("Max voltage in vm_pu:")
    for _, r in net.res_bus.sort_values("vm_pu", ascending=False).iloc[:numv].iterrows():
        logger.info("  %s at busidx %s (%s)", r.vm_pu, r.name, net.bus.name.at[r.name])
    logger.info("Min voltage in vm_pu:")
    for _, r in net.res_bus.sort_values("vm_pu").iloc[:numv].iterrows():
        logger.info("  %s at busidx %s (%s)", r.vm_pu, r.name, net.bus.name.at[r.name])
    logger.info("Max loading trafo in %:")
    if net.res_trafo is not None:
        for _, r in net.res_trafo.sort_values("loading_percent", ascending=False).iloc[
                    :numi].iterrows():
            logger.info("  %s loading at trafo %s (%s)", r.loading_percent, r.name,
                        net.trafo.name.at[r.name])
    logger.info("Max loading line in %:")
    for _, r in net.res_line.sort_values("loading_percent", ascending=False).iloc[:numi].iterrows():
        logger.info("  %s loading at line %s (%s)", r.loading_percent, r.name,
                    net.line.name.at[r.name])


def opf_task(net, delta_pq=1e-3, keep=False, log=True):
    """
    Collects some basic inforamtion of the optimal powerflow task und prints them.
    """
    if keep:
        net = copy.deepcopy(net)
    _check_necessary_opf_parameters(net, logger)

    opf_task_overview = {"flexibilities": dict(),
                         "network_constraints": dict(),
                         "flexibilities_without_costs": dict()}
    _determine_flexibilities_dict(net, opf_task_overview["flexibilities"], delta_pq)
    _determine_network_constraints_dict(net, opf_task_overview["network_constraints"])
    _determine_costs_dict(net, opf_task_overview)

    _check_overlapping_constraints(opf_task_overview)
    if log:
        _log_opf_task_overview(opf_task_overview)

    return opf_task_overview


def _determine_flexibilities_dict(net, data, delta_pq, **kwargs):
    """
    Determines which flexibilities exists in the net.

    INPUT:
        **net** - panpdapower net

        **data** (dict) - to store flexibilities information

        **delta_pq** (float) - if (abs(max - min) <= delta_pq) the variable is not assumed as
        flexible, since the range is as small as delta_pq (should be small, too).

    OPTIONAL:
        **kwargs**** - for comparing constraint columns with numpy.isclose(): rtol and atol
    """
    flex_elements = ["ext_grid", "gen", "dcline", "sgen", "load", "storage"]
    flex_tuple = tuple(zip(flex_elements, [True] * 3 + [False] * 3))

    for elm, controllable_default in flex_tuple:
        for power_type in ["P", "Q"]:
            key = power_type + elm
            if elm != "dcline":
                constraints = {"P": ["min_p_mw", "max_p_mw"],
                               "Q": ["min_q_mvar", "max_q_mvar"]}[power_type]
            else:
                constraints = {"P": ["max_p_mw"],
                               "Q": ["min_q_from_mvar", "max_q_from_mvar",
                                     "min_q_to_mvar", "max_q_to_mvar"]}[power_type]

            # determine indices of controllable elements, continue if no controllable element exists
            if elm in ["ext_grid", "dcline"]:
                controllables = net[elm].index
            elif "controllable" in net[elm].columns:
                controllables = net[elm].index[net[elm].controllable]
            elif controllable_default and net[elm].shape[0]:
                controllables = net[elm].index
            else:
                continue
            if not len(controllables):
                continue

            # consider delta_pq
            if len(constraints) >= 2 and pd.Series(constraints[:2]).isin(net[elm].columns).all():
                controllables = _find_idx_without_numerical_difference(
                    net[elm], constraints[0], constraints[1], delta_pq, idx=controllables,
                    equal_nan=False)
            if elm == "dcline" and power_type == "Q" and len(controllables) and \
                    pd.Series(constraints[2:4]).isin(net[elm].columns).all():
                controllables = _find_idx_without_numerical_difference(
                    net[elm], constraints[2], constraints[3], delta_pq, idx=controllables,
                    equal_nan=False)

            # add missing constraint columns
            for col_to_add in set(constraints) - set(net[elm].columns):
                net[elm][col_to_add] = np.nan

            data[key] = _cluster_same_floats(net[elm].loc[controllables], constraints, **kwargs)
            shorted = [col[:3] if col[:3] in ["min", "max"] else col for col in data[key].columns]
            if len(shorted) == len(set(shorted)):
                data[key].columns = shorted


def _find_idx_without_numerical_difference(df, column1, column2, delta, idx=None, equal_nan=False):
    """
    Returns indices where comlumn1 and column2 have a numerical difference bigger than delta.

    INPUT:
        **df** (DataFrame)

        **column1** (str) - name of first column within df to compare.
        The values of df[column1] must be numericals.

        **column2** (str) - name of second column within df to compare.
        The values of df[column2] must be numericals.

        **delta** (numerical) - value which defines whether indices are returned or not

    OPTIONAL:
        **idx** (iterable, None) - list of indices which should be considered only

        **equal_nan** (bool, False) - if False, indices are included where at least one value in
        df[column1] and df[column2] is NaN

    OUTPUT:
        **index** (pandas.Index) - index within idx where df[column1] and df[column2] deviates by
        at least delta or, if equal_na is True, one value is NaN
    """
    idx = idx if idx is not None else df.index
    idx_isnull = df.index[df[[column1, column2]].isnull().any(axis=1)]
    idx_without_null = idx.difference(idx_isnull)
    idx_no_delta = idx_without_null[(df.loc[idx_without_null, column1] - df.loc[
        idx_without_null, column2]).abs().values <= delta]

    if equal_nan:
        return idx_without_null.difference(idx_no_delta)
    else:
        return idx.difference(idx_no_delta)


def _determine_network_constraints_dict(net, data, **kwargs):
    """
    Determines which flexibilities exists in the net.

    INPUT:
        **net** - panpdapower net

        **data** (dict) - to store constraints information

    OPTIONAL:
        **kwargs**** - for comparing constraint columns with numpy.isclose(): rtol and atol
    """

    const_tuple = [("VMbus", "bus", ["min_vm_pu", "max_vm_pu"]),
                   ("LOADINGline", "line", ["max_loading_percent"]),
                   ("LOADINGtrafo", "trafo", ["max_loading_percent"]),
                   ("LOADINGtrafo3w", "trafo3w", ["max_loading_percent"])
                   ]
    for key, elm, constraints in const_tuple:
        missing_columns = set(constraints) - set(net[elm].columns)
        if net[elm].shape[0] and len(missing_columns) != len(constraints):

            # add missing constraint columns
            for col_to_add in missing_columns:
                net[elm][col_to_add] = np.nan

            data[key] = _cluster_same_floats(net[elm], constraints, **kwargs)
            shorted = [col[:3] if col[:3] in ["min", "max"] else col for col in data[key].columns]
            if len(shorted) == len(set(shorted)):
                data[key].columns = shorted


def _determine_costs_dict(net, opf_task_overview):
    """
    Determines which flexibilities do not have costs in the net. Each element is considered as one,
    i.e. if ext_grid 0, for instance,  is flexible in both, P and Q, and has one cost entry for P,
    it is not considered as 'flexibilities_without_costs'.

    INPUT:
        **net** - panpdapower net

        **opf_task_overview** (dict of dicts) - both, "flexibilities_without_costs" and
        "flexibilities" must be in opf_task_overview.keys()
    """

    cost_dfs = [df for df in ["poly_cost", "pwl_cost"] if net[df].shape[0]]
    if not len(cost_dfs):
        opf_task_overview["flexibilities_without_costs"] = "all"
        return

    flex_elements = ["ext_grid", "gen", "sgen", "load", "dcline", "storage"]

    for flex_element in flex_elements:

        # determine keys of opf_task_overview["flexibilities"] ending with flex_element
        keys = [power_type + flex_element for power_type in ["P", "Q"] if (
                power_type + flex_element) in opf_task_overview["flexibilities"].keys()]

        # determine indices of all flexibles
        idx_without_cost = set()
        for key in keys:
            idx_without_cost |= set(chain(*opf_task_overview["flexibilities"][key]["index"]))
            # simple alternative without itertools.chain():
        #            idx_without_cost |= {idx for idxs in opf_task_overview["flexibilities"][key][
        #                "index"] for idx in idxs}

        for cost_df in cost_dfs:
            idx_with_cost = set(net[cost_df].element[net[cost_df].et == flex_element].astype(np.int64))
            if len(idx_with_cost - idx_without_cost):
                logger.warning("These " + flex_element + "s have cost data but aren't flexible or" +
                               " have both, poly_cost and pwl_cost: " +
                               str(sorted(idx_with_cost - idx_without_cost)))
            idx_without_cost -= idx_with_cost

        if len(idx_without_cost):
            opf_task_overview["flexibilities_without_costs"][flex_element] = list(idx_without_cost)


def _cluster_same_floats(df, subset=None, **kwargs):
    """
    Clusters indices with close values. The values of df[subset] must be numericals.

    INPUT:
        **df** (DataFrame)

    OPTIONAL:
        **subset** (iterable, None) - list of columns of df which should be considered to cluster

        **kwargs**** - for numpy.isclose(): rtol and atol

    OUTPUT:
        **cluster_df** (DataFrame) - table of clustered values and corresponding lists of indices
    """
    if df.index.duplicated().any():
        logger.error("There are duplicated indices in df. Clusters will be determined but remain " +
                     "ambiguous.")
    subset = subset if subset is not None else df.select_dtypes(include=[
        np.number]).columns.tolist()
    uniq = ~df.duplicated(subset=subset).values

    # prepare cluster_df
    cluster_df = pd.DataFrame(np.empty((sum(uniq), len(subset) + 1)), columns=["index"] + subset)
    cluster_df["index"] = cluster_df["index"].astype(object)
    cluster_df[subset] = df.loc[uniq, subset].values

    if sum(uniq) == df.shape[0]:  # fast return if df has no duplicates
        for i1, idx in enumerate(df.index):
            cluster_df.at[i1, "index"] = [idx]
    else:  # determine index clusters
        i2 = 0
        for i1, uni in enumerate(uniq):
            if uni:
                cluster_df.at[i2, "index"] = list(df.index[np.isclose(
                    df[subset].values.astype(float),
                    df[subset].iloc[[i1]].values.astype(float),
                    equal_nan=True, **kwargs).all(axis=1)])
                i2 += 1

    return cluster_df


def _check_overlapping_constraints(opf_task_overview):
    """
    Logs variables where the minimum constraint is bigger than the maximum constraint.
    """
    overlap = []
    for dict_key in ["flexibilities", "network_constraints"]:
        for key, df in opf_task_overview[dict_key].items():
            min_col = [col for col in df.columns if "min" in col]
            max_col = [col for col in df.columns if "max" in col]
            n_col = min(len(min_col), len(max_col))
            for i_col in range(n_col):
                assert min_col[i_col].replace("min", "") == max_col[i_col].replace("max", "")
                if (df[min_col[i_col]] > df[max_col[i_col]]).any():
                    overlap.append(key)
    if len(overlap):
        logger.error("At these variables, there is a minimum constraint exceeding the maximum " +
                     "constraint value: " + str(overlap))


def _log_opf_task_overview(opf_task_overview):
    """
    Logs OPF task information.
    """
    s = ""
    for dict_key, data in opf_task_overview.items():
        if isinstance(data, str):
            assert dict_key == "flexibilities_without_costs"
            s += "\n\n%s flexibilities without costs" % data
            continue
        else:
            assert isinstance(data, dict)
        heading_logged = False
        keys, elms = _get_keys_and_elements_from_opf_task_dict(data)
        for key, elm in zip(keys, elms):
            assert elm in key
            df = data[key]

            if dict_key in ["flexibilities", "network_constraints"]:
                if not df.shape[0]:
                    continue
                if not heading_logged:
                    s += "\n\n%s:" % dict_key
                    heading_logged = True

                # --- logging information
                len_idx = len(list(chain(*df["index"])))
                if df.shape[0] > 1:
                    s += "\n    %ix %s" % (len_idx, key)
                else:
                    if not len(set(df.columns).symmetric_difference({"index", "min", "max"})):
                        s += "\n    %g <= %ix %s (all) <= %g" % (
                            df.loc[0, "min"], len_idx, key, df.loc[0, "max"])
                    else:
                        s += "\n    %ix %s (all) with these constraints:" % (len_idx, key)
                        for col in set(df.columns) - {"index"}:
                            s += " %s=%g" % (col, df.loc[0, col])
            elif dict_key == "flexibilities_without_costs":
                if not heading_logged:
                    s += "\n\n%s:" % dict_key
                    heading_logged = True
                s += "\n%ix %s" % (len(df), key)
            else:
                raise NotImplementedError("Key %s is unknown to this code." % dict_key)
    logger.info(s + "\n")


def _get_keys_and_elements_from_opf_task_dict(dict_):
    keys = list(dict_.keys())
    elms = ["".join(c for c in key if not c.isupper()) for key in keys]
    keys = list(np.array(keys)[np.argsort(elms)])
    elms = sorted(elms)
    return keys, elms


def switch_info(net, sidx):  # pragma: no cover
    """
    Prints what buses and elements are connected by a certain switch.
    """
    switch_type = net.switch.at[sidx, "et"]
    bidx = net.switch.at[sidx, "bus"]
    bus_name = net.bus.at[bidx, "name"]
    eidx = net.switch.at[sidx, "element"]
    if switch_type == "b":
        bus2_name = net.bus.at[eidx, "name"]
        logger.info("Switch %u connects bus %u (%s) with bus %u (%s)" % (sidx, bidx, bus_name,
                                                                         eidx, bus2_name))
    elif switch_type == "l":
        line_name = net.line.at[eidx, "name"]
        logger.info("Switch %u connects bus %u (%s) with line %u (%s)" % (sidx, bidx, bus_name,
                                                                          eidx, line_name))
    elif switch_type == "t":
        trafo_name = net.trafo.at[eidx, "name"]
        logger.info("Switch %u connects bus %u (%s) with trafo %u (%s)" % (sidx, bidx, bus_name,
                                                                           eidx, trafo_name))


def overloaded_lines(net, max_load=100):
    """
    Returns the results for all lines with loading_percent > max_load or None, if
    there are none.
    """
    if net.converged:
        return net["res_line"].index[net["res_line"]["loading_percent"] > max_load]
    else:
        raise UserWarning("The last loadflow terminated erratically, results are invalid!")


def violated_buses(net, min_vm_pu, max_vm_pu):
    """
    Returns all bus indices where vm_pu is not within min_vm_pu and max_vm_pu or returns None, if
    there are none of those buses.
    """
    if net.converged:
        return net["bus"].index[(net["res_bus"]["vm_pu"] < min_vm_pu) |
                                (net["res_bus"]["vm_pu"] > max_vm_pu)]
    else:
        raise UserWarning("The last loadflow terminated erratically, results are invalid!")


def clear_result_tables(net):
    """
    Clears all ``res_`` DataFrames in net.
    """
    for key in net.keys():
        if isinstance(net[key], pd.DataFrame) and key[:3] == "res" and net[key].shape[0]:
            net[key] = net[key].drop(net[key].index)


def res_power_columns(element_type, side=0):
    """Returns columns names of result tables for active and reactive power

    Parameters
    ----------
    element_type : str
        name of element table, e.g. "gen"
    side : typing.Union[int, str], optional
        Defines for branch elements which branch side is considered, by default 0

    Returns
    -------
    list[str]
        columns names of result tables for active and reactive power

    Examples
    --------
    >>> res_power_columns("gen")
    ["p_mw", "q_mvar"]
    >>> res_power_columns("line", "from")
    ["p_from_mw", "q_from_mvar"]
    >>> res_power_columns("line", 0)
    ["p_from_mw", "q_from_mvar"]
    >>> res_power_columns("line", "all")
    ["p_from_mw", "q_from_mvar", "p_to_mw", "q_to_mvar"]
    """
    if element_type in pp_elements(branch_elements=False, other_elements=False):
        return ["p_mw", "q_mvar"]
    elif element_type in pp_elements(bus=False, bus_elements=False, other_elements=False):
        if isinstance(side, int):
            if element_type == "trafo":
                side_options = {0: "hv", 1: "lv"}
            elif element_type == "trafo3w":
                side_options = {0: "hv", 1: "mv", 2: "lv"}
            else:
                side_options = {0: "from", 1: "to"}
            side = side_options[side]
        if side != "all":
            return [f"p_{side}_mw", f"q_{side}_mvar"]
        else:
            cols = res_power_columns(element_type, side=0) + \
                res_power_columns(element_type, side=1)
            if element_type == "trafo3w":
                cols += res_power_columns(element_type, side=2)
            return cols
    else:
        raise ValueError(f'{element_type=} cannot be considered by res_power_columns().')


import logging
import pandas as pd

from pandapower.control.controller.trafo.ContinuousTapControl import ContinuousTapControl
from pandapower.control.controller.trafo.DiscreteTapControl import DiscreteTapControl
from pandapower.converter.cim import cim_tools
from pandapower.converter.cim.cim2pp import build_pp_net

logger = logging.getLogger('cim.cim2pp.converter_classes.tapController')

sc = cim_tools.get_pp_net_special_columns_dict()


class TapController:
    def __init__(self, cimConverter: build_pp_net.CimConverter):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cimConverter = cimConverter

    def create_tap_controller_for_power_transformers(self):
        if self.cimConverter.power_trafo2w.index.size > 0:
            self._create_tap_controller(self.cimConverter.power_trafo2w, 'trafo')
        if self.cimConverter.power_trafo3w.index.size > 0:
            self._create_tap_controller(self.cimConverter.power_trafo3w, 'trafo3w')

    def _create_tap_controller(self, input_df: pd.DataFrame, trafo_type: str):
        if not self.cimConverter.kwargs.get('create_tap_controller', True):
            self.logger.info("Skip creating transformer tap changer controller for transformer type %s." % trafo_type)
            return
        for _, row in input_df.loc[input_df['TapChangerControl'].notna()].iterrows():
            trafo_id = self.cimConverter.net[trafo_type].loc[
                self.cimConverter.net[trafo_type][sc['o_id']] == row[sc['o_id']]].index.values[0]
            # get the controlled bus (side), assume "lv" as default
            side = 'lv'
            if sc['t_hv'] in self.cimConverter.net[trafo_type].columns and \
                    row['c_Terminal'] in self.cimConverter.net[trafo_type][sc['t_hv']].values:
                side = 'hv'
            if sc['t_mv'] in self.cimConverter.net[trafo_type].columns and \
                    row['c_Terminal'] in self.cimConverter.net[trafo_type][sc['t_mv']].values:
                side = 'mv'
            if row['discrete']:
                self.logger.info("Creating DiscreteTapControl for transformer %s." % row[sc['o_id']])
                DiscreteTapControl(self.cimConverter.net, element=trafo_type, element_index=trafo_id, side=side,
                                   tol=row['c_tol'], in_service=row['c_in_service'],
                                   vm_lower_pu=row['c_vm_lower_pu'], vm_upper_pu=row['c_vm_upper_pu'])
            else:
                self.logger.info("Creating ContinuousTapControl for transformer %s." % row[sc['o_id']])
                ContinuousTapControl(self.cimConverter.net, element=trafo_type, element_index=trafo_id, side=side,
                                     tol=row['c_tol'], in_service=row['c_in_service'], vm_set_pu=row['c_vm_set_pu'])



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np
import matplotlib.pyplot as plt


def plot_pq_area(pq_area, **kwargs):
    """
    Assumption: there is no q restriction at vm = 1 pu
    """
    p_samples = kwargs.pop("p_samples", None)
    if p_samples is None:
        p_samples = np.linspace(0, 1, 200)

    min_max_q = pq_area.q_flexibility(p_pu=p_samples, vm_pu=1)

    _plot_pq_area(min_max_q, p_samples=p_samples, **kwargs)


def _plot_pq_area(min_max_q, title=None, ax=None, saturate_sn_pu:float=np.nan, circle_segment=90,
                  p_samples=None, tex=False):
    if p_samples is None:
        p_samples = np.linspace(0, 1, 200)
    if tex:
        texts = "$Q_{min}$", "$Q_{max}$", \
            r"underexcited $\leftarrow------ Q/S_N ------\rightarrow$ overexcited", "$P/S_N$", \
            '$S_{max}$'
    else:
        texts = "Qmin", "Qmax", "underexcited < ----- Q/Sn ----- > overexcited", "P/Sn", 'Smax'
    if ax is None:
        plt.figure()
        ax = plt.gca()
    ax.plot(min_max_q[:, 0], p_samples, c="blue", alpha=0.9, label=texts[0])
    ax.plot(min_max_q[:, 1], p_samples, c="red", alpha=0.9, label=texts[1])
    ax.set_xlabel(texts[2])
    ax.set_ylabel(texts[3])
    ax.legend()
    ax.grid(alpha=0.8)
    if title is not None:
        ax.set_title(title)
    ax.set_ylim(p_samples.min(), p_samples.max())

    if not np.isnan(saturate_sn_pu):
        # the output x, y of generate_circle_segment() is used in reverse, since a segment around
        # 0 degree => east is used instead of around north
        y, x = generate_circle_segment(0, 0, saturate_sn_pu, -circle_segment, circle_segment, 1)
        ax.plot(x, y, '--k', label=texts[4])
        y_circ = np.sqrt(saturate_sn_pu**2 - p_samples**2)
        x_diff = max(x) - min(x)
        ax.set_xlim(min(x)-0.05*x_diff, max(x)+0.05*x_diff)
    else:
        y_circ = np.ones(p_samples.shape)
        x_diff = max(min_max_q[:, 1]) - min(min_max_q[:, 0])
        ax.set_xlim(min(min_max_q[:, 0])-0.05*x_diff, max(min_max_q[:, 1])+0.05*x_diff)
    ax.fill_betweenx(p_samples, np.max(np.c_[min_max_q[:, 0], -y_circ], axis=1),
                     np.min(np.c_[min_max_q[:, 1], y_circ], axis=1), color="green", alpha=0.2)
    plt.tight_layout()


def plot_qv_area(qv_area, title=None, ax=None, prune_to_flexibility=False, vm_samples=None, tex=False):
    if vm_samples is None:
        vm_samples = np.linspace(0.8, 1.2, 200)
    if tex:
        texts = "$Q_{min}$", "$Q_{max}$", "$v_m$ in pu", "$Q/S_N$"
    else:
        texts = "Qmin", "Qmax", "Vm in pu", "Q/Sn"

    min_max_q = qv_area.q_flexibility(p_pu=0.5, vm_pu=vm_samples) # p should be irrelevant since
    # it is about QV areas, not about PQV areas

    len_ = min_max_q.shape[0]

    if prune_to_flexibility:
        start = min(np.arange(len_, dtype=int)[~np.isclose(min_max_q[:, 0], min_max_q[0, 0])])-1
        stop = max(np.arange(len_, dtype=int)[~np.isclose(min_max_q[:, 1], min_max_q[-1, 1])])+2
        vm_samples = vm_samples[start:stop]
        min_max_q = min_max_q[start:stop, :]

    # --- do plotting
    if ax is None:
        plt.figure()
        ax = plt.gca()

    ax.plot(vm_samples, min_max_q[:, 0], c="blue", alpha=0.9, label=texts[0])
    ax.plot(vm_samples, min_max_q[:, 1], c="red", alpha=0.9, label=texts[1])
    ax.set_xlabel(texts[2])
    ax.set_ylabel(texts[3])
    ax.legend()
    ax.grid(alpha=0.8)
    ax.fill_between(vm_samples, min_max_q[:, 0], min_max_q[:, 1], color="green", alpha=0.2)

    if title is not None:
        ax.set_title(title)
    plt.tight_layout()


def generate_circle_segment(center_x, center_y, radius, start, stop, step):
    """
    generates coordinates for a semicircle, centered at center_x, center_y
    generates x and y coordinates for a segment of a circle, centered at center_x, center_y. The
    segment (start, stop, step) is given in degree.
    Which degree means which direction?
    0: "east", 90: "north", 180: "west", 270: "south"

    Example
    -------
    >>> x, y = generate_circle_segment(0.5, 0.7, 1, 90, 225, 1)
    >>> plt.plot(x, y)
    >>> plt.axis('equal')
    >>> plt.show()
    Even a helix can be generated:
    >>> x, y = generate_circle_segment(0, 0, np.linspace(0, 1, 3*360+1), 0, 3*360, 1)
    >>> plt.plot(x, y)
    >>> plt.axis('equal')
    >>> plt.show()
    """
    degrees = np.array([start, stop + step/2, step])
    angles = np.radians(degrees)
    sample_angles = np.arange(angles[0], angles[1], angles[2])
    x = center_x + radius * np.cos(sample_angles)
    y = center_y + radius * np.sin(sample_angles)
    return x, y


if __name__ == "__main__":
    from pandapower.control.controller.DERController import PQVArea4120V1, PQVArea4120V2, PQVArea4120V3

    pq_area = PQVArea4120V1()
    plot_pq_area(pq_area, "PQVArea4120V1")
    plt.show()
    plot_qv_area(pq_area, "PQVArea4120V1")
    plt.show()

    pq_area = PQVArea4120V2()
    plot_pq_area(pq_area, "PQVArea4120V2")
    plt.show()
    plot_qv_area(pq_area, "PQVArea4120V2")
    plt.show()

    pq_area = PQVArea4120V3()
    plot_pq_area(pq_area, "PQVArea4120V3")
    plt.show()
    plot_qv_area(pq_area, "PQVArea4120V3")
    plt.show()



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from pandapower.pypower.idx_brch import branch_cols as start

# todo: add comments to describe the variables!

IKSS_F    = start + 0
IKSS_T    = start + 1
ITH_F     = start + 2
ITH_T     = start + 3
IP_F      = start + 4
IP_T      = start + 5
IK_F      = start + 6
IK_T      = start + 7
IB_F      = start + 8
IB_T      = start + 9
K_T       = start + 10
K_ST       = start + 11


PKSS_F    = start + 12
QKSS_F    = start + 13
PKSS_T    = start + 14
QKSS_T    = start + 15
VKSS_MAGN_F    = start + 16
VKSS_MAGN_T    = start + 17
VKSS_ANGLE_F = start + 18
VKSS_ANGLE_T = start + 19
IKSS_ANGLE_F = start + 20
IKSS_ANGLE_T = start + 21

branch_cols_sc = 22



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from numpy import nan_to_num, array, allclose, int64

from pandapower.auxiliary import LoadflowNotConverged, AlgorithmUnknown, _clean_up, _add_auxiliary_elements
from pandapower.build_branch import _calc_trafo_parameter, _calc_trafo3w_parameter
from pandapower.build_gen import _build_gen_ppc
from pandapower.pd2ppc import _pd2ppc, _calc_pq_elements_and_add_on_ppc, _ppc2ppci
from pandapower.pf.ppci_variables import _get_pf_variables_from_ppci
from pandapower.pf.run_bfswpf import _run_bfswpf
from pandapower.pf.run_dc_pf import _run_dc_pf
from pandapower.pf.run_newton_raphson_pf import _run_newton_raphson_pf
from pandapower.pf.runpf_pypower import _runpf_pypower
from pandapower.pypower.bustypes import bustypes
from pandapower.pypower.idx_bus import VM
from pandapower.pypower.makeYbus import makeYbus as makeYbus_pypower
from pandapower.pypower.pfsoln import pfsoln as pfsoln_pypower
from pandapower.results import _extract_results, _copy_results_ppci_to_ppc, init_results, \
    verify_results, _ppci_bus_to_ppc, _ppci_other_to_ppc

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def _powerflow(net, **kwargs):
    """
    Gets called by runpp or rundcpp with different arguments.
    """

    # get infos from options
    ac = net["_options"]["ac"]
    algorithm = net["_options"]["algorithm"]

    net["converged"] = False
    net["OPF_converged"] = False
    _add_auxiliary_elements(net)  # create gen elements for start and end buses of dcline

    if not ac or net["_options"]["init_results"]:
        verify_results(net)
    else:
        init_results(net)

    if net["_options"]["voltage_depend_loads"] and algorithm not in ['nr', 'bfsw'] and not (
            allclose(net.load.const_z_percent.values, 0) and
            allclose(net.load.const_i_percent.values, 0)):
        logger.error(("pandapower powerflow does not support voltage depend loads for algorithm "
                      "'%s'!") % algorithm)

    # clear lookups
    net._pd2ppc_lookups = {"bus": array([], dtype=int64), "bus_dc": array([], dtype=int64),
                           "ext_grid": array([], dtype=int64), "gen": array([], dtype=int64),
                           "branch": array([], dtype=int64), "branch_dc": array([], dtype=int64)}

    # convert pandapower net to ppc
    ppc, ppci = _pd2ppc(net, **kwargs)

    # store variables
    net["_ppc"] = ppc

    if "VERBOSE" not in kwargs:
        kwargs["VERBOSE"] = 0

    # ----- run the powerflow -----
    result = _run_pf_algorithm(ppci, net["_options"], **kwargs)
    # read the results (=ppci with results) to net
    _ppci_to_net(result, net)


def _recycled_powerflow(net, **kwargs):
    options = net["_options"]
    options["recycle"] = kwargs.get("recycle", None)
    options["init_vm_pu"] = "results"
    options["init_va_degree"] = "results"
    algorithm = options["algorithm"]
    ac = options["ac"]
    recycle = options["recycle"]
    ppci = {"bus": net["_ppc"]["internal"]["bus"],
            "gen": net["_ppc"]["internal"]["gen"],
            "branch": net["_ppc"]["internal"]["branch"],
            "baseMVA": net["_ppc"]["internal"]["baseMVA"],
            "internal": net["_ppc"]["internal"],
            }
    if algorithm not in ['nr', 'iwamoto_nr'] and ac:
        raise ValueError("recycle is only available with Newton-Raphson power flow. Choose "
                         "algorithm='nr'")

    ppc = net["_ppc"]
    ppc["success"] = False
    ppc["iterations"] = 0.
    ppc["et"] = 0.

    if "bus_pq" in recycle and recycle["bus_pq"]:
        # update pq values in bus
        _calc_pq_elements_and_add_on_ppc(net, ppc)

    if "trafo" in recycle and recycle["trafo"]:
        # update trafo in branch and Ybus
        lookup = net._pd2ppc_lookups["branch"]
        if "trafo" in lookup:
            _calc_trafo_parameter(net, ppc,)
        if "trafo3w" in lookup:
            _calc_trafo3w_parameter(net, ppc)

    if "gen" in recycle and recycle["gen"]:
        # updates the ppc["gen"] part
        _build_gen_ppc(net, ppc)
        ppc["gen"] = nan_to_num(ppc["gen"])

    ppci = _ppc2ppci(ppc, net, ppci=ppci)
    ppci["internal"] = net["_ppc"]["internal"]
    net["_ppc"] = ppc

    if not ac:
        # DC recycle
        result = _run_dc_pf(ppci, recycle)
        _ppci_to_net(result, net)
        return

    # run the Newton-Raphson power flow
    result = _run_newton_raphson_pf(ppci, options)
    ppc["success"] = ppci["success"]
    ppc["iterations"] = ppci["iterations"]
    ppc["et"] = ppci["et"]
    if options["only_v_results"]:
        _ppci_bus_to_ppc(result, ppc)
        _ppci_other_to_ppc(result, ppc, options["mode"])
        return
    # read the results from  result (==ppci) to net
    _ppci_to_net(result, net)


def _run_pf_algorithm(ppci, options, **kwargs):
    algorithm = options["algorithm"]
    ac = options["ac"]

    if ac:
        _, pv, pq = bustypes(ppci["bus"], ppci["gen"])
        # ----- run the powerflow -----
        if pq.shape[0] == 0 and pv.shape[0] == 0 and not options['distributed_slack'] \
                and len(ppci["svc"]) == 0 and len(ppci["tcsc"]) == 0 and len(ppci["ssc"]) == 0 and len(ppci["vsc"]) == 0:
            # ommission not correct if distributed slack is used or facts devices are present
            result = _bypass_pf_and_set_results(ppci, options)
        elif algorithm == 'bfsw':  # forward/backward sweep power flow algorithm
            result = _run_bfswpf(ppci, options, **kwargs)[0]
        elif algorithm in ['nr', 'iwamoto_nr']:
            result = _run_newton_raphson_pf(ppci, options)
        elif algorithm in ['fdbx', 'fdxb', 'gs']:  # algorithms existing within pypower
            result = _runpf_pypower(ppci, options, **kwargs)[0]
        else:
            raise AlgorithmUnknown("Algorithm {0} is unknown!".format(algorithm))
    else:
        result = _run_dc_pf(ppci, options["recycle"])

    return result


def _ppci_to_net(result, net):
    # reads the results from result (== ppci with results) to pandapower net

    mode = net["_options"]["mode"]
    # ppci doesn't contain out of service elements, but ppc does -> copy results accordingly
    ppc = net["_ppc"]
    result = _copy_results_ppci_to_ppc(result, ppc, mode)

    # raise if PF was not successful. If DC -> success is always 1
    if result["success"] != 1:
        _clean_up(net, res=False)
        algorithm = net["_options"]["algorithm"]
        max_iteration = net["_options"]["max_iteration"]
        raise LoadflowNotConverged("Power Flow {0} did not converge after "
                                   "{1} iterations!".format(algorithm, max_iteration))
    else:
        net["_ppc"] = result
        net["converged"] = True

    _extract_results(net, result)
    _clean_up(net)


def _bypass_pf_and_set_results(ppci, options):
    Ybus, Yf, Yt = makeYbus_pypower(ppci["baseMVA"], ppci["bus"], ppci["branch"])
    baseMVA, bus, gen, branch, svc, tcsc, ssc, vsc, ref, _, pq, *_, V0, ref_gens = _get_pf_variables_from_ppci(ppci)
    V = ppci["bus"][:, VM]
    bus, gen, branch = pfsoln_pypower(baseMVA, bus, gen, branch, svc, tcsc, ssc, vsc, Ybus, Yf, Yt, V, ref, ref_gens)
    ppci["bus"], ppci["gen"], ppci["branch"] = bus, gen, branch
    ppci["success"] = True
    ppci["iterations"] = 1
    ppci["et"] = 0
    return ppci



from pandapower.plotting.collections import *
from pandapower.plotting.colormaps import *
from pandapower.plotting.generic_geodata import *
from pandapower.plotting.powerflow_results import *
from pandapower.plotting.simple_plot import *
from pandapower.plotting.plotly import *
from pandapower.plotting.geo import *
from pandapower.plotting.plotting_toolbox import set_line_geodata_from_bus_geodata
from pandapower.plotting.to_html import to_html

import types

try:
    from matplotlib.backend_bases import GraphicsContextBase, RendererBase

    class GC(GraphicsContextBase):
        def __init__(self):
            super().__init__()
            self.set_capstyle('round')

    def custom_new_gc(self):
        return GC()

    RendererBase.new_gc = types.MethodType(custom_new_gc, RendererBase)

except ImportError:
    pass
